import atexit
import logging
import weakref
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Type,
    TypeVar,
    Union,
)

from weave.trace.context import get_raise_on_captured_errors
from weave.trace.op import FinishCallbackType, Op
from weave.trace.op_extensions.log_once import log_once

logger = logging.getLogger(__name__)

S = TypeVar("S")
V = TypeVar("V")


_OnYieldType = Callable[[V], None]
_OnErrorType = Callable[[Exception], None]
_OnCloseType = Callable[[], None]


class _IteratorWrapper(Generic[V]):
    """This class wraps an iterator object allowing hooks to be added to the lifecycle of the iterator. It is likely
    that this class will be helpful in other contexts and might be moved to a more general location in the future."""

    def __init__(
        self,
        iterator_or_ctx_manager: Union[Iterator, AsyncIterator],
        on_yield: _OnYieldType,
        on_error: _OnErrorType,
        on_close: _OnCloseType,
    ) -> None:
        self._iterator_or_ctx_manager = iterator_or_ctx_manager
        self._on_yield = on_yield
        self._on_error = on_error
        self._on_close = on_close
        self._on_finished_called = False

        atexit.register(weakref.WeakMethod(self._call_on_close_once))

    def _call_on_close_once(self) -> None:
        if not self._on_finished_called:
            try:
                self._on_close()  # type: ignore
            except Exception as e:
                log_once(
                    logger.error,
                    f"Error finishing call - some logs may not be captured: {e}",
                )
                if get_raise_on_captured_errors():
                    raise
            self._on_finished_called = True

    def _call_on_error_once(self, e: Exception) -> None:
        if not self._on_finished_called:
            try:
                self._on_error(e)
            except Exception as e:
                log_once(
                    logger.error,
                    f"Error finishing call with exception - some logs may not be captured: {e}",
                )
                if get_raise_on_captured_errors():
                    raise
            self._on_finished_called = True

    def __iter__(self) -> "_IteratorWrapper":
        return self

    def __next__(self) -> Generator[None, None, V]:
        if not hasattr(self._iterator_or_ctx_manager, "__next__"):
            raise TypeError(
                f"Cannot call next on an iterator of type {type(self._iterator_or_ctx_manager)}"
            )
        try:
            value = next(self._iterator_or_ctx_manager)  # type: ignore
            try:
                # Here we do a try/catch because we don't want to
                # break the user process if we trip up on processing
                # the yielded value
                self._on_yield(value)
            except Exception as e:
                log_once(
                    logger.error, f"Error capturing yielded value for call output: {e}"
                )
                if get_raise_on_captured_errors():
                    raise
            return value
        except (StopIteration, StopAsyncIteration) as e:
            self._call_on_close_once()
            raise
        except Exception as e:
            self._call_on_error_once(e)
            raise

    def __aiter__(self) -> "_IteratorWrapper":
        return self

    async def __anext__(self) -> Generator[None, None, V]:
        if not hasattr(self._iterator_or_ctx_manager, "__anext__"):
            raise TypeError(
                f"Cannot call anext on an iterator of type {type(self._iterator_or_ctx_manager)}"
            )
        try:
            value = await self._iterator_or_ctx_manager.__anext__()  # type: ignore
            try:
                self._on_yield(value)
                # Here we do a try/catch because we don't want to
                # break the user process if we trip up on processing
                # the yielded value
            except Exception as e:
                log_once(
                    logger.error,
                    f"Error capturing async yielded value for call output: {e}",
                )
                if get_raise_on_captured_errors():
                    raise
            return value
        except (StopAsyncIteration, StopIteration) as e:
            self._call_on_close_once()
            raise StopAsyncIteration
        except Exception as e:
            self._call_on_error_once(e)
            raise

    def __del__(self) -> None:
        self._call_on_close_once()

    def close(self) -> None:
        self._call_on_close_once()

    def __getattr__(self, name: str) -> Any:
        """Delegate all other attributes to the wrapped iterator."""
        if name in [
            "_iterator_or_ctx_manager",
            "_on_yield",
            "_on_error",
            "_on_close",
            "_on_finished_called",
            "_call_on_error_once",
        ]:
            return object.__getattribute__(self, name)
        return getattr(self._iterator_or_ctx_manager, name)

    def __enter__(self) -> "_IteratorWrapper":
        if hasattr(self._iterator_or_ctx_manager, "__enter__"):
            # let's enter the context manager to get the stream iterator
            self._iterator_or_ctx_manager = self._iterator_or_ctx_manager.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Exception],
        exc_value: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        if exc_type and isinstance(exc_value, Exception):
            self._call_on_error_once(exc_value)
        if hasattr(
            self._iterator_or_ctx_manager, "__exit__"
        ):  # case where is a context mngr
            self._iterator_or_ctx_manager.__exit__(exc_type, exc_value, traceback)
        self._call_on_close_once()

    async def __aenter__(self) -> "_IteratorWrapper":
        if hasattr(
            self._iterator_or_ctx_manager, "__aenter__"
        ):  # let's enter the context manager
            self._iterator_or_ctx_manager = (
                await self._iterator_or_ctx_manager.__aenter__()
            )
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Exception],
        exc_value: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        if exc_type and isinstance(exc_value, Exception):
            self._call_on_error_once(exc_value)
        self._call_on_close_once()


def add_accumulator(
    op: Op,
    make_accumulator: Callable[[Dict], Callable[[S, V], S]],
    *,
    should_accumulate: Optional[Callable[[Dict], bool]] = None,
    on_finish_post_processor: Optional[Callable[[Any], Any]] = None,
    iterator_wrapper: Type[_IteratorWrapper] = _IteratorWrapper,
) -> Op:
    """This is to be used internally only - specifically designed for integrations with streaming libraries.

    Add an accumulator to an op. The accumulator will be called with the output of the op
    after the op is resolved. The accumulator should return the output of the op. This is intended
    for internal use only and may change in the future. The accumulator should take two arguments:
    the current state of the accumulator and the value to accumulate. It should return the new state
    of the accumulator. The first time the accumulator is called, the current state will be None.

    The intended usage is:

    ```
    @weave.op()
    def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    def simple_list_accumulator(acc, value):
        if acc is None:
            acc = []
        acc.append(value)
        return acc
    add_accumulator(fn, simple_list_accumulator) # returns the op with `list(range(9, -1, -1))` as output
    """

    def on_output(
        value: Iterator[V], on_finish: FinishCallbackType, inputs: Dict
    ) -> Iterator:
        def wrapped_on_finish(value: Any, e: Optional[BaseException] = None) -> None:
            if on_finish_post_processor is not None:
                value = on_finish_post_processor(value)
            on_finish(value, e)

        if should_accumulate is None or should_accumulate(inputs):
            # we build the accumulator here dependent on the inputs (optional)
            accumulator = make_accumulator(inputs)
            return _build_iterator_from_accumulator_for_op(
                value,
                accumulator,
                wrapped_on_finish,
                iterator_wrapper,
            )
        else:
            wrapped_on_finish(value)
            return value

    op._set_on_output_handler(on_output)
    return op


def _build_iterator_from_accumulator_for_op(
    value: Iterator[V],
    accumulator: Callable,
    on_finish: FinishCallbackType,
    iterator_wrapper: Type["_IteratorWrapper"] = _IteratorWrapper,
) -> "_IteratorWrapper":
    acc: _Accumulator = _Accumulator(accumulator)

    def on_yield(value: V) -> None:
        acc.next(value)

    def on_error(e: Exception) -> None:
        on_finish(acc.get_state(), e)

    def on_close() -> None:
        on_finish(acc.get_state(), None)

    return iterator_wrapper(value, on_yield, on_error, on_close)


class _Accumulator(Generic[S, V]):
    state: Optional[S]

    def __init__(
        self,
        accumulator: Callable[[Optional[S], V], S],
        initial_state: Optional[S] = None,
    ):
        self._accumulator = accumulator
        self._state = initial_state

    def next(self, value: V) -> None:
        # the try-except hack to catch `StopIteration` inside `<integration>_accumulator`
        # this `StopIteration` is raised when some condition is met, for example, when
        # we don't want to surface last chunk (with usage info) from openai integration.
        try:
            self._state = self._accumulator(self._state, value)
        except StopIteration as e:
            self._state = e.value
            raise

    def get_state(self) -> Optional[S]:
        return self._state



import pytest

from weave.legacy.weave.util import relpath_no_syscalls


@pytest.mark.parametrize(
    "target_path, start_path, current_working_directory, expected",
    [
        # First batch
        (
            "/home/user/documents/report.txt",
            "/home/user/projects",
            "/home/user",
            "../documents/report.txt",
        ),
        ("reports/summary.pdf", "downloads", "/home/user", "../reports/summary.pdf"),
        ("/home/user/documents", "/home/user/documents", "/home/user", "."),
        # Second batch (mixed cases)
        (
            "/home/user/documents/report.txt",
            "projects",
            "/home/user",
            "../documents/report.txt",
        ),
        ("/var/log", "downloads", "/home/user", "../../../var/log"),
        (
            "projects/subproject/code.py",
            "/usr/local",
            "/home/user",
            "../../home/user/projects/subproject/code.py",
        ),
        # Edge cases
        ("/", "/", "/", "."),
        ("/", "subfolder", "/home/user", "../../.."),
        ("/home/user/subfolder", ".", "/home/user", "subfolder"),
        ("/home/user/subfolder", "..", "/home/user/subfolder", "subfolder"),
        (
            "/home/user/documents/file.txt",
            "../../../documents",
            "/home/user/projects/subproject",
            "../user/documents/file.txt",
        ),
        (
            "/home/user/documents/file.txt",
            "../../documents",
            "/home/user/projects/subproject",
            "file.txt",
        ),
        ("./file.txt", ".", "/home/user", "file.txt"),
        ("./file.txt", "./", "/home/user", "file.txt"),
        ("././file.txt", "./", "/home/user", "file.txt"),
    ],
)
def test_relpath_no_syscalls(
    target_path, start_path, current_working_directory, expected
):
    assert (
        relpath_no_syscalls(target_path, start_path, current_working_directory)
        == expected
    )



import contextlib
import os
import time

import pytest

import weave
from weave.legacy.weave.monitoring import monitor

### Skipping some tests here. They are for features that no longer exist. Since we're
# iterating on the API, I'm not removing them yet.


@contextlib.contextmanager
def async_disabled():
    current = os.environ.get("WEAVE_DISABLE_ASYNC_FILE_STREAM")
    os.environ["WEAVE_DISABLE_ASYNC_FILE_STREAM"] = "true"
    try:
        yield
    finally:
        if current is None:
            del os.environ["WEAVE_DISABLE_ASYNC_FILE_STREAM"]
        else:
            os.environ["WEAVE_DISABLE_ASYNC_FILE_STREAM"] = current


@pytest.mark.skip()
def test_monitoring_basic(user_by_api_key_in_env):
    with async_disabled():
        mon = monitor.new_monitor(
            "%s/%s/test_monitoring" % (user_by_api_key_in_env.username, "test")
        )

        @mon.trace()
        def example(a, b):
            return a + b

        rows = []
        results = []
        for i in range(10):
            for j in range(10):
                example(i, j)
                rows.append({"a": i, "b": j})
                results.append(i + j)

        # These underscore accesses are not public API. This is just for testing.
        mon._streamtable.finish()
        node = mon.rows()
        inputs_node = node["inputs"]
        output_node = node["output"]
        input_results, output_results = weave.use([inputs_node, output_node])

        assert input_results.to_pylist_tagged() == rows
        assert output_results.to_pylist_tagged() == results


@pytest.mark.skip("Feature removed")
def test_monitoring_auto_false(user_by_api_key_in_env):
    with async_disabled():
        # @monitoring.monitor(
        #     entity_name=user_by_api_key_in_env.username,
        #     project_name="test",
        #     auto_log=False,
        # )
        def example(a, b):
            return a + b

        rows = []
        results = []
        for i in range(10):
            for j in range(10):
                res = example(i, j)
                res.add_data({"c": i + j})
                results.append(res)
                rows.append(i + j)

        for res in results[::-1]:
            # Do the finishing in reverse order
            res.finalize()

        # These underscore accesses are not public API. This is just for testing.
        example._stream_table.finish()
        node = example.rows()["c"]

        assert weave.use(node).to_pylist_tagged() == rows[::-1]


@pytest.mark.skip("Feature removed")
def test_monitoring_capture_errors(user_by_api_key_in_env):
    with async_disabled():
        # @monitoring.monitor(
        #     entity_name=user_by_api_key_in_env.username,
        #     project_name="test",
        #     raise_on_error=False,
        # )
        def example(a, b):
            if b == 5:
                raise ValueError("5 is bad")
            return a + b

        rows = []
        results = []
        exceptions = []
        for i in range(10):
            for j in range(10):
                example(i, j)
                rows.append({"a": i, "b": j})
                if j == 5:
                    results.append(None)
                    exceptions.append("ValueError: 5 is bad")
                else:
                    results.append(i + j)
                    exceptions.append(None)

        # These underscore accesses are not public API. This is just for testing.
        example._stream_table.finish()
        node = example.rows()

        input_results, output_results, exception_results = weave.use(
            [node["inputs"], node["output"], node["exception"]]
        )

        assert input_results.to_pylist_tagged() == rows
        assert output_results.to_pylist_tagged() == results
        assert exception_results.to_pylist_tagged() == exceptions


@pytest.mark.skip("Feature removed")
def test_monitoring_processors(user_by_api_key_in_env):
    with async_disabled():
        # @monitoring.monitor(
        #     entity_name=user_by_api_key_in_env.username,
        #     project_name="test",
        #     input_preprocessor=lambda a, b: {"val": f"{a} + {b}"},
        #     output_postprocessor=lambda res: {"val": res},
        # )
        def example(a, b):
            return a + b

        rows = []
        results = []
        for i in range(10):
            for j in range(10):
                example(i, j)
                rows.append({"val": f"{i} + {j}"})
                results.append({"val": i + j})

        # These underscore accesses are not public API. This is just for testing.
        example._stream_table.finish()
        node = example.rows()

        input_results, output_results = weave.use([node["inputs"], node["output"]])

        assert input_results.to_pylist_tagged() == rows
        assert output_results.to_pylist_tagged() == results



from typing import Callable

from weave.flow.obj import Object

INFER_METHOD_NAMES = {"predict", "infer", "forward", "invoke"}


class MissingInferenceMethodError(Exception): ...


class Model(Object):
    """
    Intended to capture a combination of code and data the operates on an input.
    For example it might call an LLM with a prompt to make a prediction or generate
    text.

    When you change the attributes or the code that defines your model, these changes
    will be logged and the version will be updated. This ensures that you can compare
    the predictions across different versions of your model. Use this to iterate on
    prompts or to try the latest LLM and compare predictions across different settings

    Examples:

    ```python
    class YourModel(Model):
        attribute1: str
        attribute2: int

        @weave.op()
        def predict(self, input_data: str) -> dict:
            # Model logic goes here
            prediction = self.attribute1 + ' ' + input_data
            return {'pred': prediction}
    ```
    """

    # TODO: should be infer: Callable

    def get_infer_method(self) -> Callable:
        for name in INFER_METHOD_NAMES:
            if infer_method := getattr(self, name, None):
                return infer_method
        raise MissingInferenceMethodError(
            f"Missing a method with name in ({INFER_METHOD_NAMES})"
        )


def get_infer_method(model: Model) -> Callable:
    for name in INFER_METHOD_NAMES:
        if (infer_method := getattr(model, name, None)) is not None:
            return infer_method
    raise MissingInferenceMethodError(
        f"Missing a method with name in ({INFER_METHOD_NAMES})"
    )



import dataclasses
import typing
from urllib import parse

from weave.legacy.weave import box
from weave.trace_server import refs_internal

DICT_KEY_EDGE_NAME = refs_internal.DICT_KEY_EDGE_NAME
LIST_INDEX_EDGE_NAME = refs_internal.LIST_INDEX_EDGE_NAME
OBJECT_ATTR_EDGE_NAME = refs_internal.OBJECT_ATTR_EDGE_NAME
AWL_ROW_EDGE_NAME = "row"
AWL_COL_EDGE_NAME = "col"


def parse_local_ref_str(s: str) -> typing.Tuple[str, typing.Optional[list[str]]]:
    if "#" not in s:
        return s, None
    path, extra = s.split("#", 1)
    return path, extra.split("/")


def val_with_relative_ref(
    parent_object: typing.Any, child_object: typing.Any, ref_extra_parts: list[str]
) -> typing.Any:
    from weave.legacy.weave import context_state, ref_base

    # If we already have a ref, resolve it
    if isinstance(child_object, ref_base.Ref):
        child_object = child_object.get()

    # Only do this if ref_tracking_enabled right now. I just want to
    # avoid introducing new behavior into W&B prod for the moment.
    if context_state.ref_tracking_enabled():
        from weave.legacy.weave import storage

        child_ref = storage.get_ref(child_object)
        parent_ref = ref_base.get_ref(parent_object)

        # This first check is super important - if the child ref is pointing
        # to a completely different artifact (ref), then we want to point to
        # the child's inherent ref, not the relative ref from the parent.
        if child_ref is not None:
            if parent_ref is not None:
                if hasattr(child_ref, "digest") and hasattr(parent_ref, "digest"):
                    if child_ref.digest != parent_ref.digest:
                        return child_object

        if parent_ref is not None:
            child_object = box.box(child_object)
            sub_ref = parent_ref.with_extra(None, child_object, ref_extra_parts)
            ref_base._put_ref(child_object, sub_ref)
        return child_object

    return child_object


@dataclasses.dataclass
class RefExtraTuple:
    edge_type: str
    part: str


@dataclasses.dataclass
class ParsedRef:
    scheme: str
    entity: typing.Optional[str]
    project: typing.Optional[str]
    artifact: str
    alias: str
    file_path_parts: list[str]
    ref_extra_tuples: list[RefExtraTuple]


def parse_ref_str(s: str) -> ParsedRef:
    scheme, _, path, _, _, ref_extra = parse.urlparse(s)
    entity = None
    project = None
    assert path.startswith("/")
    path = path[1:]
    path_parts = path.split("/")
    if scheme == "wandb-artifact":
        entity = path_parts[0]
        project = path_parts[1]
        path_parts = path_parts[2:]

    artifact, alias = path_parts[0].split(":")
    file_path_parts = path_parts[1:]
    ref_extra_tuples = []
    if ref_extra:
        ref_extra_parts = ref_extra.split("/")
        assert len(ref_extra_parts) % 2 == 0
        for i in range(0, len(ref_extra_parts), 2):
            edge_type = ref_extra_parts[i]
            part = ref_extra_parts[i + 1]
            ref_extra_tuples.append(RefExtraTuple(edge_type, part))

    return ParsedRef(
        scheme=scheme,
        entity=entity,
        project=project,
        artifact=artifact,
        alias=alias,
        file_path_parts=file_path_parts,
        ref_extra_tuples=ref_extra_tuples,
    )



import json
import typing

from weave.legacy.weave import api as weave
from weave.legacy.weave import file_base


def load_jsonl(jsonlfile):
    rows = []
    for row in jsonlfile:
        rows.append(json.loads(row))
    return rows


@weave.op(
    name="file-refine_readjsonl",
    hidden=True,
    input_type={"self": file_base.FileBaseType(extension=weave.types.literal("jsonl"))},
)
def refine_readjsonl(self) -> weave.types.Type:
    with self.open() as f:
        return weave.types.TypeRegistry.type_of(load_jsonl(f))


@weave.op(
    name="file-readjsonl",
    input_type={"self": file_base.FileBaseType(extension=weave.types.literal("jsonl"))},
    output_type=weave.types.List(weave.types.TypedDict({})),
    refine_output_type=refine_readjsonl,
)
def readjsonl(self):
    with self.open() as f:
        return load_jsonl(f)


@weave.op()
def json_dumps(data: typing.Any) -> str:
    return json.dumps(data)



import weave
from weave.legacy.weave import weave_internal


def test_weave_fn_in_data():
    weave_fn = weave_internal.define_fn({"x": weave.types.Int()}, lambda x: x + 3)
    data = weave.save({"a": weave_fn})
    assert weave.use(data["a"](5)) == 8


def test_weave_fn_in_data_called_in_callback():
    weave_fn = weave_internal.define_fn({"x": weave.types.Int()}, lambda x: x + 3)
    data = weave.save({"a": weave_fn})
    my_list = weave.save([1, 2, 3])
    assert weave.use(my_list.filter(lambda row: data["a"](row) > 4)) == [2, 3]



import math

import pyarrow as pa

from weave.legacy.weave import mappers_arrow
from weave.legacy.weave import weave_types as types


def test_map_list():
    d = [{"a": 5, "b": "x"}]
    d_type = types.TypeRegistry.type_of(d)
    m = mappers_arrow.map_to_arrow(d_type, None)
    res_type = m.result_type()
    assert res_type == pa.list_(
        pa.field(
            "item", pa.struct([pa.field("a", pa.int64()), pa.field("b", pa.string())])
        )
    )
    d2 = m.apply(d)
    assert d == d2
    m2 = mappers_arrow.map_from_arrow(d_type, None)
    d3 = m2.apply(d2)
    assert d2 == d3


def test_map_list_with_none():
    d = [{"a": 5, "b": None}]
    d_type = types.TypeRegistry.type_of(d)
    m = mappers_arrow.map_to_arrow(d_type, None)
    res_type = m.result_type()
    assert res_type == pa.list_(
        pa.field(
            "item", pa.struct([pa.field("a", pa.int64()), pa.field("b", pa.null())])
        )
    )
    d2 = m.apply(d)
    assert d == d2
    m2 = mappers_arrow.map_from_arrow(d_type, None)
    d3 = m2.apply(d2)
    assert d2 == d3


def test_map_list_with_maybe():
    d = [{"a": 6, "b": "x"}, {"a": 5, "b": None}]
    d_type = types.TypeRegistry.type_of(d)
    m = mappers_arrow.map_to_arrow(d_type, None)
    res_type = m.result_type()
    assert res_type == pa.list_(
        pa.field(
            "item", pa.struct([pa.field("a", pa.int64()), pa.field("b", pa.string())])
        )
    )
    d2 = m.apply(d)
    assert d == d2
    m2 = mappers_arrow.map_from_arrow(d_type, None)
    d3 = m2.apply(d2)
    assert d2 == d3


def test_map_list_with_nan():
    d = [{"a": 5, "b": float("nan")}]
    d_type = types.TypeRegistry.type_of(d)
    m = mappers_arrow.map_to_arrow(d_type, None)
    res_type = m.result_type()
    assert res_type == pa.list_(
        pa.field(
            "item", pa.struct([pa.field("a", pa.int64()), pa.field("b", pa.float64())])
        )
    )
    d2 = m.apply(d)
    assert d[0]["a"] == d2[0]["a"]
    assert math.isnan(d2[0]["b"])
    m2 = mappers_arrow.map_from_arrow(d_type, None)
    d3 = m2.apply(d2)
    assert d2[0]["a"] == d3[0]["a"]
    assert math.isnan(d3[0]["b"])


def test_mapper_list_to_arrow_arr_has_no_type():
    obj = [
        {
            "IDs": [264, 268, 29, 77, 47, 297, 73, 18, 366, 364],
            "Transformations": "a",
        },
        {
            "IDs": None,
            "Transformations": "b",
        },
    ]
    ot = types.TypeRegistry.type_of(obj)

    # this should not raise an exception
    m = mappers_arrow.map_to_arrow(ot, None)
    m2 = mappers_arrow.map_from_arrow(ot, None)
    transformed = m.apply(obj)
    untransformed = m2.apply(transformed)

    # this should work with nulls
    assert untransformed == obj



from typing import Any, TypeVar

from typing_extensions import TypeGuard

from weave.trace.object_record import ObjectRecord
from weave.trace.vals import WeaveObject

C = TypeVar("C")


def weave_isinstance(obj: Any, cls: type[C]) -> TypeGuard[C]:
    if isinstance(obj, cls):  # type: ignore
        return True
    if isinstance(obj, ObjectRecord):
        return obj._class_name == cls.__name__ or any(
            b == cls.__name__ for b in obj._bases
        )
    if isinstance(obj, WeaveObject):
        return obj._class_name == cls.__name__ or any(
            b == cls.__name__ for b in obj._bases
        )
    return False



import dataclasses
import typing

import pyarrow as pa

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import artifact_base, box
from weave.legacy.weave.arrow import arrow_tags, convert
from weave.legacy.weave.arrow.arrow import ArrowWeaveListType, arrow_as_array
from weave.legacy.weave.arrow.list_ import ArrowWeaveList
from weave.legacy.weave.language_features.tagging import tag_store, tagged_value_type


def repeat(value: typing.Any, count: int) -> pa.Array:
    value_single = convert.to_arrow([value])._arrow_data
    if count == 0:
        return pa.array([], value_single.type)
    return pa.repeat(value_single[0], count)


@dataclasses.dataclass
class VectorizedContainerConstructorResults:
    arrays: list[pa.Array]
    prop_types: dict[str, types.Type]
    max_len: int
    artifact: typing.Optional[artifact_base.Artifact]


def vectorized_input_types(input_types: dict[str, types.Type]) -> dict[str, types.Type]:
    prop_types: dict[str, types.Type] = {}
    for input_name, input_type in input_types.items():
        if isinstance(input_type, types.Const):
            input_type = input_type.val_type
        if isinstance(input_type, tagged_value_type.TaggedValueType) and (
            isinstance(input_type.value, ArrowWeaveListType)
            or types.is_list_like(input_type.value)
        ):
            outer_tag_type = input_type.tag
            object_type = input_type.value.object_type  # type: ignore
            if isinstance(object_type, tagged_value_type.TaggedValueType):
                new_prop_type = tagged_value_type.TaggedValueType(
                    types.TypedDict(
                        {
                            **outer_tag_type.property_types,
                            **object_type.tag.property_types,
                        }
                    ),
                    object_type.value,
                )
            else:
                new_prop_type = tagged_value_type.TaggedValueType(
                    outer_tag_type, object_type
                )
            prop_types[input_name] = new_prop_type
        elif isinstance(input_type, ArrowWeaveListType) or types.is_list_like(
            input_type
        ):
            prop_types[input_name] = input_type.object_type  # type: ignore
        else:  # is scalar
            prop_types[input_name] = input_type
    return prop_types


def vectorized_container_constructor_preprocessor(
    input_dict: dict[str, typing.Any],
) -> VectorizedContainerConstructorResults:
    if len(input_dict) == 0:
        return VectorizedContainerConstructorResults([], {}, 0, None)

    arrays = []
    prop_types = {}
    awl_artifact = None
    for k, v in input_dict.items():
        if isinstance(v, ArrowWeaveList):
            if awl_artifact is None:
                awl_artifact = v._artifact
            if tag_store.is_tagged(v):
                list_tags = tag_store.get_tags(v)
                # convert tags to arrow
                v = arrow_tags.tag_awl_list_elements_with_single_tag_dict(v, list_tags)
            prop_types[k] = v.object_type
            v = v._arrow_data
            arrays.append(arrow_as_array(v))
        else:
            prop_types[k] = types.TypeRegistry.type_of(v)
            arrays.append(v)

    # array len of None means we have a scalar
    array_lens: list[typing.Optional[int]] = []
    for a, t in zip(arrays, prop_types.values()):
        if hasattr(a, "to_pylist"):
            array_lens.append(len(a))
        else:
            array_lens.append(None)

    if all(l is None for l in array_lens):
        max_len = 1
    else:
        max_len = max(a for a in array_lens if a is not None)

    for l in array_lens:
        if l is not None and l != max_len:
            raise errors.WeaveInternalError(
                f"Cannot create ArrowWeaveDict with different length arrays (scalars are ok): {array_lens}"
            )

    for i, (a, l) in enumerate(zip(arrays, array_lens)):
        if l is None:
            tags: typing.Optional[dict] = None
            if tag_store.is_tagged(a):
                tags = tag_store.get_tags(a)
            if box.is_boxed(a):
                a = box.unbox(a)
            arrays[i] = repeat(a, max_len)
            if tags is not None:
                arrays[i] = arrow_tags.tag_arrow_array_elements_with_single_tag_dict(
                    arrays[i], tags
                )

    return VectorizedContainerConstructorResults(
        arrays, prop_types, max_len, awl_artifact
    )



__all__ = ["patch", "unpatch"]

import functools
import typing
from contextlib import contextmanager
from functools import partialmethod
from typing import AsyncIterator, Callable, Type, Union

import openai
from openai import AsyncStream, Stream
from openai.types.chat import ChatCompletion, ChatCompletionMessageParam
from packaging import version
from weave.trace import call_context
from weave.trace.client_context import weave_client as weave_client_context
from weave.legacy.weave.monitoring.monitor import _get_global_monitor
from weave.legacy.weave.monitoring.openai.models import *
from weave.legacy.weave.monitoring.openai.util import *
from weave.trace.op import Op
from weave.trace.op import op as op_deco

old_create = openai.resources.chat.completions.Completions.create
old_async_create = openai.resources.chat.completions.AsyncCompletions.create

create_op_name = "openai.chat.completions.create"
create_op: typing.Union[str, Op] = create_op_name
try:
    create_op = op_deco(old_create)
    create_op.name = create_op_name
except Exception as e:
    pass


def to_python(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {k: to_python(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [to_python(v) for v in obj]
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    return obj


class partialmethod_with_self(partialmethod):
    def __get__(self, obj: Any, cls: Optional[Type[Any]] = None) -> Callable:
        return self._make_unbound_method().__get__(obj, cls)  # type: ignore


class WeaveAsyncStream(AsyncStream):
    def __init__(
        self,
        *,
        base_stream: AsyncStream,
        messages: List[ChatCompletionMessageParam],
        finish_run: Callable,
    ) -> None:
        self._messages = messages
        self._chunks: List[ChatCompletionChunk] = []
        self._finish_run = finish_run
        super().__init__(
            cast_to=ChatCompletionChunk,
            client=base_stream._client,
            response=base_stream.response,
        )

    async def __anext__(self) -> ChatCompletionChunk:
        try:
            item = await self._iterator.__anext__()
        except StopAsyncIteration:
            self._on_finish()
            raise
        self._chunks.append(item)
        return item

    async def __aiter__(self) -> AsyncIterator[ChatCompletionChunk]:
        async for item in self._iterator:
            self._chunks.append(item)
            yield item

        self._on_finish()

    def _on_finish(self) -> None:
        from weave.flow.chat_util import OpenAIStream

        wrapped_stream = OpenAIStream(iter(self._chunks))
        list(wrapped_stream)

        result = wrapped_stream.final_response()
        result_with_usage = ChatCompletion(
            **result.model_dump(exclude_unset=True),
            usage=token_usage(self._messages, result.choices),
        )
        self._finish_run(result_with_usage.model_dump(exclude_unset=True))


class WeaveStream(Stream):
    def __init__(
        self,
        *,
        base_stream: Stream,
        messages: List[ChatCompletionMessageParam],
        finish_run: Callable,
    ) -> None:
        from weave.flow.chat_util import OpenAIStream

        self._messages = messages
        self._chunks: List[ChatCompletionChunk] = []
        self._finish_run = finish_run
        super().__init__(
            cast_to=ChatCompletionChunk,
            client=base_stream._client,
            response=base_stream.response,
        )
        self._openai_stream = OpenAIStream(self._iterator)

    def __next__(self) -> ChatCompletionChunk:
        try:
            return self._openai_stream.__next__()
        except StopIteration:
            self._on_finish()
            raise

    def __iter__(self) -> Iterator[ChatCompletionChunk]:
        for chunk in self._openai_stream:
            yield chunk
        self._on_finish()

    def _on_finish(self) -> None:
        result = self._openai_stream.final_response()
        result_with_usage = ChatCompletion(
            **result.model_dump(exclude_unset=True),
            usage=token_usage(self._messages, result.choices),
        )
        self._finish_run(result_with_usage.model_dump(exclude_unset=True))


class AsyncChatCompletions:
    def __init__(self, base_create: Callable) -> None:
        self._base_create = base_create

    async def create(
        self, *args: Any, **kwargs: Any
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        if kwargs.get("stream", False):
            return await self._streaming_create(*args, **kwargs)
        return await self._create(*args, **kwargs)

    async def _create(self, *args: Any, **kwargs: Any) -> ChatCompletion:
        named_args = bind_params(old_create, *args, **kwargs)
        with log_call(create_op, named_args) as finish_run:
            result = await self._base_create(*args, **kwargs)
            finish_run(result.model_dump(exclude_unset=True))
        return result

    async def _streaming_create(
        self, *args: Any, **kwargs: Any
    ) -> AsyncStream[ChatCompletionChunk]:
        from weave.flow.chat_util import OpenAIStream

        named_args = bind_params(old_async_create, *args, **kwargs)
        messages = to_python(named_args["messages"])
        if not isinstance(messages, list):
            raise ValueError("messages must be a list")
        with log_call(create_op, named_args) as finish_run:
            # We return a special AsyncStream that mimics the underlying
            # one, but also logs the result of the completion.
            base_stream = await self._base_create(*args, **kwargs)
            stream = WeaveAsyncStream(
                base_stream=base_stream,
                messages=messages,
                finish_run=finish_run,
            )
        return stream


class ChatCompletions:
    def __init__(self, base_create: Callable) -> None:
        self._base_create = base_create

    def create(
        self, *args: Any, **kwargs: Any
    ) -> Union[ChatCompletion, Stream[ChatCompletionChunk]]:
        if kwargs.get("stream", False):
            result = self._streaming_create(*args, **kwargs)
            return result
        return self._create(*args, **kwargs)

    def _create(self, *args: Any, **kwargs: Any) -> ChatCompletion:
        named_args = bind_params(old_create, *args, **kwargs)
        with log_call(create_op, named_args) as finish_run:
            result = self._base_create(*args, **kwargs)
            finish_run(result.model_dump(exclude_unset=True))
        return result

    def _streaming_create(
        self, *args: Any, **kwargs: Any
    ) -> Stream[ChatCompletionChunk]:
        named_args = bind_params(old_create, *args, **kwargs)
        messages = to_python(named_args["messages"])
        if not isinstance(messages, list):
            raise ValueError("messages must be a list")

        with log_call(create_op, named_args) as finish_run:
            base_stream = self._base_create(*args, **kwargs)
            stream = WeaveStream(
                base_stream=base_stream,
                messages=messages,
                finish_run=finish_run,
            )

        return stream


def patch() -> None:
    def _patch() -> None:
        unpatch_fqn = f"{unpatch.__module__}.{unpatch.__qualname__}()"

        gc = weave_client_context.require_weave_client()
        if gc:
            # info(f"Patching OpenAI completions.  To unpatch, call {unpatch_fqn}")

            hooks = ChatCompletions(old_create)
            async_hooks = AsyncChatCompletions(old_async_create)
            openai.resources.chat.completions.Completions.create = (
                partialmethod_with_self(hooks.create)
            )
            openai.resources.chat.completions.AsyncCompletions.create = (
                partialmethod_with_self(async_hooks.create)
            )
        else:
            error("No graph client found, not patching OpenAI completions")

    if version.parse(openai.__version__) < version.parse("1.0.0"):
        error(
            f"this integration requires openai>=1.0.0 (got {openai.__version__}).  Please upgrade and try again"
        )
        return

    try:
        _patch()
    except Exception as e:
        error(f"problem patching: {e}, auto-unpatching")
        unpatch()
        raise Exception from e


def unpatch() -> None:
    # if _get_global_monitor() is not None:
    # info("Unpatching OpenAI completions")
    openai.resources.chat.completions.Completions.create = old_create
    openai.resources.chat.completions.AsyncCompletions.create = old_async_create


# TODO: centralize
@contextmanager
def log_call(
    call_name: typing.Union[str, Op], inputs: dict[str, Any]
) -> Iterator[Callable]:
    client = weave_client_context.require_weave_client()
    parent_call = call_context.get_current_call()
    # TODO: client should not need refs passed in.
    call = client.create_call(call_name, inputs, parent_call)

    def finish_call(output: Any) -> None:
        # TODO: client should not need refs passed in.
        client.finish_call(call, output)

    try:
        yield finish_call
    except Exception as e:
        client.fail_call(call, e)
        raise



import pytest

from weave.trace_server import validation
from weave.trace_server.errors import InvalidRequest


def test_validate_dict_one_key():
    with pytest.raises(InvalidRequest) as e:
        validation.validate_dict_one_key("foo", "foo", str)
    assert str(e.value) == "Expected a dictionary, got foo"

    with pytest.raises(InvalidRequest) as e:
        validation.validate_dict_one_key({}, "foo", str)
    assert str(e.value) == "Expected a dictionary with one key, got {}"

    with pytest.raises(InvalidRequest) as e:
        validation.validate_dict_one_key({"foo": "bar"}, "bar", str)
    assert str(e.value) == "Expected key bar, got foo"

    with pytest.raises(InvalidRequest) as e:
        validation.validate_dict_one_key({"foo": 12}, "foo", str)
    assert str(e.value) == "Expected value of type <class 'str'>, got <class 'int'>"

    assert validation.validate_dict_one_key({"foo": "bar"}, "foo", str) == "bar"


def test_validate_purge_req_one():
    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one("foo")
    assert str(e.value) == "Expected a dictionary, got foo"

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one({"foo": "bar"})
    assert str(e.value) == "Expected key eq_, got foo"

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one({"eq_": "bar"})
    assert str(e.value) == "Expected value of type <class 'tuple'>, got <class 'str'>"

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one({"eq_": {"get_field_": 12}})
    assert str(e.value) == "Expected value of type <class 'tuple'>, got <class 'dict'>"

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one({"eq_": tuple([{"get_field_": "id"}])})
    assert str(e.value) == validation.MESSAGE_INVALID_PURGE

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one(
            {"eq_": {"get_field_": "foo"}, "literal_": "bar"}
        )
    assert (
        str(e.value)
        == "Expected a dictionary with one key, got {'eq_': {'get_field_': 'foo'}, 'literal_': 'bar'}"
    )

    validation.validate_purge_req_one(
        {"eq_": tuple([{"get_field_": "id"}, {"literal_": "bar"}])}
    )

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one({"in_": ()}, operator="in_")
    assert str(e.value) == validation.MESSAGE_INVALID_PURGE

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one(
            {"in_": tuple({"literal_": 12})}, operator="in_"
        )
    assert str(e.value) == validation.MESSAGE_INVALID_PURGE

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_one(
            {"in_": tuple({"literal_": "bar"})}, operator="in_"
        )
    assert str(e.value) == validation.MESSAGE_INVALID_PURGE

    validation.validate_purge_req_one(
        {
            "in_": tuple(
                [
                    {"get_field_": "id"},
                    [{"literal_": "foo"}],
                ]
            )
        },
        operator="in_",
    )
    validation.validate_purge_req_one(
        {
            "in_": tuple(
                [
                    {"get_field_": "id"},
                    [{"literal_": "bar"}, {"literal_": "foo"}],
                ]
            )
        },
        operator="in_",
    )


def test_validate_purge_req_multiple():
    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_multiple("foo")
    assert str(e.value) == validation.MESSAGE_INVALID_PURGE

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_multiple({"foo": "bar"})
    assert str(e.value) == validation.MESSAGE_INVALID_PURGE

    with pytest.raises(InvalidRequest) as e:
        validation.validate_purge_req_multiple(
            [{"eq_": {"get_field_": "id"}, "literal_": "bar"}]
        )
    assert (
        str(e.value)
        == "Expected a dictionary with one key, got {'eq_': {'get_field_': 'id'}, 'literal_': 'bar'}"
    )

    validation.validate_purge_req_multiple(
        [
            {"eq_": tuple([{"get_field_": "id"}, {"literal_": "bar"}])},
            {"eq_": tuple([{"get_field_": "id"}, {"literal_": "bar"}])},
        ]
    )



import weave
from weave.legacy.weave import ref_base, uris, weave_internal


def test_mutation_set_direct_call():
    val = weave.legacy.weave.ops.TypedDict.pick({"a": {"b": 5}}, "a")["b"]
    set_result = weave.legacy.weave.ops.set(val, 9)
    assert set_result == {"a": {"b": 9}}


def test_mutation_set_dispatch():
    val = weave.legacy.weave.ops.TypedDict.pick({"a": {"b": 5}}, "a")["b"]
    set_result = val.set(9)
    assert set_result == {"a": {"b": 9}}


def test_mutation_artifact():
    weave.save([1, 2, 3], "art:main")
    art = weave.legacy.weave.ops.get("local-artifact:///art:main/obj")
    art.append(4)
    new_art = weave.storage.get("local-artifact:///art:main/obj")
    assert new_art == [1, 2, 3, 4]


def test_mutation_lazy():
    # This is how weavejs does it.
    weave.save([1, 2, 3], "art:main")
    art = weave.legacy.weave.ops.get("local-artifact:///art:main/obj")

    # Quote lhs expr, so that it is not evaluated.
    quoted_art = weave_internal.const(art)

    expr = weave.legacy.weave.ops.append.lazy_call(quoted_art, 4, {})
    weave.use(expr)

    new_art = weave.storage.get("local-artifact:///art:main/obj")
    assert new_art == [1, 2, 3, 4]


def test_mutation_lazy_works_without_quoting():
    # This relies on "auto quoting behavior". See compile_quote
    weave.save([1, 2, 3], "art:main")
    art = weave.legacy.weave.ops.get("local-artifact:///art:main/obj")

    expr = weave.legacy.weave.ops.append.lazy_call(art, 4, {})
    weave.use(expr)

    new_art = weave.storage.get("local-artifact:///art:main/obj")
    assert new_art == [1, 2, 3, 4]


def test_merge():
    weave.save({"a": 5, "b": 6}, "my-dict:latest")
    dict_obj = weave.legacy.weave.ops.get("local-artifact:///my-dict:latest/obj")
    weave.legacy.weave.ops.set(dict_obj["a"], 17, root_args={"branch": "my-branch"})
    modified_dict_obj = weave.legacy.weave.ops.get("local-artifact:///my-dict:my-branch/obj")
    new_uri = weave.legacy.weave.ops.merge_artifact(modified_dict_obj)
    dict_obj_node = weave.legacy.weave.ops.get(new_uri)
    assert (
        weave.use(dict_obj_node)
        == weave.use(weave.legacy.weave.ops.get("local-artifact:///my-dict:latest/obj"))
        == {"a": 17, "b": 6}
    )


def test_merge_no_version():
    get_node = weave.save({"a": 5, "b": 6}, "my-dict")
    uri = get_node.from_op.inputs["uri"].val  # type: ignore

    # uri now has a direct commit hash for the version
    dict_obj = weave.legacy.weave.ops.get(uri)
    weave.legacy.weave.ops.set(dict_obj["a"], 17, root_args={"branch": "my-branch"})
    modified_dict_obj = weave.legacy.weave.ops.get("local-artifact:///my-dict:my-branch/obj")
    new_uri = weave.legacy.weave.ops.merge_artifact(modified_dict_obj)
    dict_obj_node = weave.legacy.weave.ops.get(new_uri)
    assert weave.use(dict_obj_node) == {"a": 17, "b": 6}


def test_merge_list_type():
    from weave.legacy.weave import object_context

    weave.save([], "my-list:latest")
    obj = weave.legacy.weave.ops.get("local-artifact:///my-list:latest/obj")
    with object_context.object_context():
        obj.append({"a": "x"}, {})
        obj.append([1], {})

    assert weave.use(weave.legacy.weave.ops.get("local-artifact:///my-list:latest/obj")) == [
        {"a": "x"},
        [1],
    ]


def test_artifact_history_local():
    num_versions = 4
    uri = "local-artifact:///art:main/obj"
    weave.save([0], "art:main")
    art = weave.legacy.weave.ops.get(uri)

    for i in range(num_versions):
        art.append(i + 1)

    total_list = list(range(num_versions + 1))
    new_art = weave.storage.get(uri)
    assert new_art == total_list

    for i in range(num_versions):
        new_uri = weave.legacy.weave.ops.undo_artifact(weave.legacy.weave.ops.get(uri))
        # We expect these to be the same since the branch pointer changed
        assert new_uri == uri
        res = weave.storage.get(uri)
        assert res == total_list[: num_versions - i]


def test_artifact_history_local_from_hash():
    num_versions = 4
    uri = "local-artifact:///art:main/obj"
    weave.save([0], "art:main")
    art = weave.legacy.weave.ops.get(uri)

    for i in range(num_versions):
        art.append(i + 1)

    hash_uri = weave.legacy.weave.uris.WeaveURI.parse(uri).to_ref().artifact.uri
    assert "main" not in hash_uri

    total_list = list(range(num_versions + 1))
    new_art = weave.storage.get(hash_uri + "/obj")
    assert new_art == total_list

    new_uri = hash_uri
    for i in range(num_versions):
        new_uri = weave.legacy.weave.ops.undo_artifact(weave.legacy.weave.ops.get(new_uri))
        assert "main" not in new_uri
        res = weave.storage.get(new_uri + "/obj")
        assert res == total_list[: num_versions - i]


def test_artifact_history_remote_with_branch(user_by_api_key_in_env):
    num_versions = 2
    uri = "local-artifact:///art:main/obj"
    weave.save([0], "art:main")
    art = weave.legacy.weave.ops.get(uri)
    published_art_uri = weave.legacy.weave.ops.publish_artifact(art, "art", None, None)

    art = weave.legacy.weave.ops.get(
        f"wandb-artifact:///{user_by_api_key_in_env.username}/weave/art:latest/obj"
    )

    for i in range(num_versions):
        res_uri = art.append(i + 1)
        art = weave.legacy.weave.ops.get(res_uri)

    new_uri = res_uri
    total_list = list(range(num_versions + 1))
    new_art = weave.storage.get(new_uri)
    assert new_art == total_list

    for i in range(num_versions):
        new_uri = weave.legacy.weave.ops.undo_artifact(weave.legacy.weave.ops.get(new_uri))
        res = weave.storage.get(new_uri)
        if i == num_versions - 1:
            assert new_uri.startswith("wandb")
        else:
            assert new_uri.startswith("local")
        assert res == total_list[: num_versions - i]


def test_artifact_history_remote_with_hash(user_by_api_key_in_env):
    num_versions = 2
    uri = "local-artifact:///art:main/obj"
    weave.save([0], "art:main")
    art = weave.legacy.weave.ops.get(uri)
    published_art_uri = weave.legacy.weave.ops.publish_artifact(art, "art", None, None)
    assert "latest" not in published_art_uri
    assert "main" not in published_art_uri

    art = weave.legacy.weave.ops.get(published_art_uri)

    for i in range(num_versions):
        res_uri = art.append(i + 1)
        art = weave.legacy.weave.ops.get(res_uri)

    new_uri = res_uri
    total_list = list(range(num_versions + 1))
    new_art = weave.storage.get(new_uri)
    assert new_art == total_list

    for i in range(num_versions):
        new_uri = weave.legacy.weave.ops.undo_artifact(weave.legacy.weave.ops.get(new_uri))
        res = weave.storage.get(new_uri)
        if i == num_versions - 1:
            assert new_uri.startswith("wandb")
        else:
            assert new_uri.startswith("local")
        assert res == total_list[: num_versions - i]



import inspect
from contextlib import contextmanager
from typing import Any, Callable, Dict, Generator, Iterator, Optional, TypeVar

import tiktoken
import wandb
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk, Choice
from openai.types.completion_usage import CompletionUsage

from weave.legacy.weave.monitoring.openai.models import *

T = TypeVar("T")


def update_combined_choice(
    combined_choice: CombinedChoice, choice: Choice
) -> CombinedChoice:
    combined_choice.content += choice.delta.content or ""
    combined_choice.role = combined_choice.role or choice.delta.role
    combined_choice.function_call = (
        combined_choice.function_call or choice.delta.function_call
    )
    combined_choice.tool_calls = combined_choice.tool_calls or choice.delta.tool_calls
    if choice.finish_reason:
        combined_choice.finish_reason = choice.finish_reason
    return combined_choice


def token_usage(
    input_messages: List[dict], response_choices: list[Choice]
) -> CompletionUsage:
    prompt_tokens = num_tokens_from_messages(input_messages)
    completion_tokens = 0
    for choice in response_choices:
        message = choice.message
        completion_tokens += num_tokens_from_messages([message.model_dump()])

    total_tokens = prompt_tokens + completion_tokens
    return CompletionUsage(
        prompt_tokens=prompt_tokens,
        completion_tokens=completion_tokens,
        total_tokens=total_tokens,
    )


def num_tokens_from_messages(
    messages: List[dict], model: str = "gpt-3.5-turbo-0613"
) -> int:
    model_defaults = {
        "gpt-3.5-turbo-0613": ModelTokensConfig(per_message=3, per_name=1),
        "gpt-3.5-turbo-16k-0613": ModelTokensConfig(per_message=3, per_name=1),
        "gpt-4-0314": ModelTokensConfig(per_message=3, per_name=1),
        "gpt-4-32k-0314": ModelTokensConfig(per_message=3, per_name=1),
        "gpt-4-0613": ModelTokensConfig(per_message=3, per_name=1),
        "gpt-4-32k-0613": ModelTokensConfig(per_message=3, per_name=1),
        "gpt-3.5-turbo-0301": ModelTokensConfig(per_message=4, per_name=-1),
    }

    config = model_defaults.get(model)
    if config is None:
        if "gpt-3.5-turbo" in model:
            print(
                "Warning: gpt-3.5-turbo may update over time. Assuming gpt-3.5-turbo-0613."
            )
            return num_tokens_from_messages(messages, "gpt-3.5-turbo-0613")
        elif "gpt-4" in model:
            print("Warning: gpt-4 may update over time. Assuming gpt-4-0613.")
            return num_tokens_from_messages(messages, "gpt-4-0613")
        else:
            raise NotImplementedError(
                f"num_tokens_from_messages() is not implemented for model {model}. "
                "See https://github.com/openai/openai-python/blob/main/chatml.md "
                "for information on how messages are converted to tokens."
            )

    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    num_tokens = 3  # Prime with assistant
    for message in messages:
        num_tokens += config.per_message
        # message['content'] is a list when we're sending images.
        # TODO: This doesn't account for image tokens, we just skip this
        # case!
        if message.get("content") is not None and not isinstance(
            message.get("content"), list
        ):
            num_tokens += len(encoding.encode(message["content"]))
        if message["role"] == "user":
            num_tokens += config.per_name

    return num_tokens


def info(msg: str) -> None:
    wandb.termlog(msg)


def error(msg: str) -> None:
    wandb.termerror(msg)


def warn(msg: str) -> None:
    wandb.termwarn(msg)


@contextmanager
def error_handler() -> Generator[None, None, None]:
    try:
        yield
    except Exception as e:
        print(f"problem with callback: {e}")


def bind_params(func: Callable, *args: Any, **kwargs: Any) -> Dict[str, Any]:
    sig = inspect.signature(func)
    return sig.bind_partial(*args, **kwargs).arguments



import contextlib
import io
import os
import tempfile
from typing import Any, Dict, Generator, Iterator, Mapping, Optional, Union

from weave.legacy.weave import artifact_fs
from weave.trace import op_type  # noqa: F401, Must import this to register op save/load
from weave.trace.client_context.weave_client import require_weave_client
from weave.trace.op import Op, op
from weave.trace.refs import ObjectRef, parse_uri
from weave.trace.serializer import get_serializer_by_id, get_serializer_for_obj


class MemTraceFilesArtifact(artifact_fs.FilesystemArtifact):
    RefClass = artifact_fs.FilesystemArtifactRef
    temp_read_dir: Optional[tempfile.TemporaryDirectory]
    path_contents: dict[str, bytes]

    def __init__(
        self,
        path_contents: Optional[Mapping[str, Union[str, bytes]]] = None,
        metadata: Optional[dict[str, str]] = None,
    ):
        if path_contents is None:
            path_contents = {}
        self.path_contents = path_contents  # type: ignore
        if metadata is None:
            metadata = {}
        self._metadata = metadata
        self.temp_read_dir = None

    @contextlib.contextmanager
    def new_file(
        self, path: str, binary: bool = False
    ) -> Iterator[Union[io.StringIO, io.BytesIO]]:
        f: Union[io.StringIO, io.BytesIO]
        if binary:
            f = io.BytesIO()
        else:
            f = io.StringIO()
        yield f
        self.path_contents[path] = f.getvalue()  # type: ignore
        f.close()

    @property
    def is_saved(self) -> bool:
        return True

    @contextlib.contextmanager
    def open(
        self, path: str, binary: bool = False
    ) -> Iterator[Union[io.StringIO, io.BytesIO]]:
        f: Union[io.StringIO, io.BytesIO]
        try:
            if binary:
                val = self.path_contents[path]
                if not isinstance(val, bytes):
                    raise ValueError(
                        f"Expected binary file, but got string for path {path}"
                    )
                f = io.BytesIO(val)
            else:
                val = self.path_contents[path]
                f = io.StringIO(val.decode("utf-8"))
        except KeyError:
            raise FileNotFoundError(path)
        yield f
        f.close()

    def path(self, path: str) -> str:
        if path not in self.path_contents:
            raise FileNotFoundError(path)

        self.temp_read_dir = tempfile.TemporaryDirectory()
        write_path = os.path.join(self.temp_read_dir.name, path)
        with open(write_path, "wb") as f:
            f.write(self.path_contents[path])
            f.flush()
            os.fsync(f.fileno())
        return write_path

    @property
    def metadata(self) -> artifact_fs.ArtifactMetadata:
        return artifact_fs.ArtifactMetadata(self._metadata, {**self._metadata})

    @contextlib.contextmanager
    def writeable_file_path(self, path: str) -> Generator[str, None, None]:
        with tempfile.TemporaryDirectory() as tmpdir:
            full_path = os.path.join(tmpdir, path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            yield full_path
            with open(full_path, "rb") as fp:
                self.path_contents[path] = fp.read()


def encode_custom_obj(obj: Any) -> Optional[dict]:
    serializer = get_serializer_for_obj(obj)
    if serializer is None:
        # We silently return None right now. We could warn here. This object
        # will not be recoverable with client.get
        return None
    art = MemTraceFilesArtifact()
    serializer.save(obj, art, "obj")

    # Save the load_instance function as an op, and store a reference
    # to that op in the saved value record. We don't do this if what
    # we're saving is actually an op, since that would be self-referential
    # (the op loading code is always present, we don't need to save/load it).
    load_op_uri = None
    if serializer.id() != "Op":
        # Ensure load_instance is an op
        if not isinstance(serializer.load, Op):
            serializer.load = op(serializer.load)
        # Save the load_intance_op
        wc = require_weave_client()

        # TODO(PR): this can fail right? Or does it return None?
        load_instance_op_ref = wc._save_op(serializer.load, "load_" + serializer.id())  # type: ignore
        load_op_uri = load_instance_op_ref.uri()

    encoded_path_contents = {
        k: (v.encode("utf-8") if isinstance(v, str) else v)  # type: ignore
        for k, v in art.path_contents.items()
    }
    return {
        "_type": "CustomWeaveType",
        "weave_type": {"type": serializer.id()},
        "files": encoded_path_contents,
        "load_op": load_op_uri,
    }


def decode_custom_obj(
    weave_type: Dict,
    encoded_path_contents: Mapping[str, Union[str, bytes]],
    load_instance_op_uri: Optional[str],
) -> Any:
    from weave.legacy.weave import artifact_fs

    load_instance_op = None
    if load_instance_op_uri is not None:
        ref = parse_uri(load_instance_op_uri)
        if not isinstance(ref, ObjectRef):
            raise ValueError(f"Expected ObjectRef, got {load_instance_op_uri}")
        wc = require_weave_client()
        load_instance_op = wc.get(ref)
        if load_instance_op is None:
            raise ValueError(
                f"Failed to load op needed to decode object of type {weave_type}. See logs above for more information."
            )

    if load_instance_op is None:
        serializer = get_serializer_by_id(weave_type["type"])
        if serializer is None:
            raise ValueError(f"No serializer found for {weave_type}")
        load_instance_op = serializer.load

    # Disables tracing so that calls to loading data itself don't get traced
    load_instance_op._tracing_enabled = False  # type: ignore

    art = MemTraceFilesArtifact(
        encoded_path_contents,
        metadata={},
    )
    with artifact_fs.loading_artifact(art):
        res = load_instance_op(art, "obj")
        res.art = art
        return res



import os

import pytest
from cerebras.cloud.sdk import AsyncCerebras, Cerebras

import weave

model = "llama3.1-8b"  # Cerebras model


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_cerebras_sync(client: weave.trace.weave_client.WeaveClient) -> None:
    api_key = os.environ.get("CEREBRAS_API_KEY", "DUMMY_API_KEY")
    cerebras_client = Cerebras(api_key=api_key)

    response = cerebras_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "What is the capital of France?"}],
    )

    exp = "The capital of France is Paris."
    assert response.choices[0].message.content.strip() == exp

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.choices[0].message.content.strip() == exp
    assert output.choices[0].finish_reason == "stop"
    assert output.id == response.id
    assert output.model == response.model
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.completion_tokens == model_usage["completion_tokens"]
    assert output.usage.prompt_tokens == model_usage["prompt_tokens"]
    assert output.usage.total_tokens == model_usage["total_tokens"]


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_cerebras_async(client: weave.trace.weave_client.WeaveClient) -> None:
    api_key = os.environ.get("CEREBRAS_API_KEY", "DUMMY_API_KEY")
    cerebras_client = AsyncCerebras(api_key=api_key)

    response = await cerebras_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "What is the capital of France?"}],
    )

    exp = "The capital of France is Paris."
    assert response.choices[0].message.content.strip() == exp

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.choices[0].message.content.strip() == exp
    assert output.choices[0].finish_reason == "stop"
    assert output.id == response.id
    assert output.model == response.model
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.completion_tokens == model_usage["completion_tokens"]
    assert output.usage.prompt_tokens == model_usage["prompt_tokens"]
    assert output.usage.total_tokens == model_usage["total_tokens"]



import pytest

from weave.legacy.weave import api, async_demo, ops, runs, storage


def test_run_basic():
    run_id = "TESTRUN"
    run = runs.Run(id=run_id, op_name="test-run")
    run_name = "run-%s:latest" % run_id
    storage.save(run, name=run_name)

    run_node = ops.get(f"local-artifact:///{run_name}/obj")
    assert api.use(run_node.state) == "pending"
    run_node.set_state("running")
    assert api.use(run_node.state) == "running"
    run_node.print("Hello")
    run_node.print("Hello again")
    run_node.log({"x": 49.0})
    run_node.set_output("some-output")

    saved_prints = api.use(run_node.prints)
    assert saved_prints == ["Hello", "Hello again"]
    saved_logs = api.use(run_node._get_prop("history"))
    assert saved_logs == [{"x": 49.0}]
    saved_output = api.use(run_node.output)
    assert saved_output == "some-output"


@pytest.mark.timeout(1.5)
def test_automatic_await():
    twelve = async_demo.slowmult(3, 4, 0.01)
    twenty_four = async_demo.slowmult(2, twelve, 0.01)
    assert api.use(twenty_four.await_final_output()) == 24


@pytest.mark.timeout(1.5)
def test_stable_when_fetching_input():
    dataset = [{"prompt": "a", "completion": "5"}]
    ref = storage.save(dataset)
    get_dataset = ops.get(ref.uri)
    # We're going to fetch a new in memory dataset object for both of these.
    # But we should get the same run ID back because the datasets are the same
    # (which we know by checking that they come from the same ref in the code)
    train1 = async_demo.train(get_dataset)
    train2 = async_demo.train(get_dataset)
    run_id1 = api.use(train1.id())
    run_id2 = api.use(train2.id())
    assert run_id1 == run_id2
    api.use(train1.await_final_output())
    api.use(train2.await_final_output())


@pytest.mark.timeout(1.5)
def test_run_ops():
    twelve = async_demo.slowmult(3, 4, 0.01)

    # We can call any ops that are available on the run's output type.
    # So this should not fail!
    api.use(twelve + 9) == 21


@pytest.mark.timeout(1.5)
def test_run_ops_mapped():
    input = api.save([1, 2])
    result = input.map(
        lambda item: async_demo.slowmult(item, 4, 0.01).await_final_output()
    )

    # We can call any ops that are available on the run's output type.
    # So this should not fail!
    api.use(result) == [5, 6]


@pytest.mark.timeout(1.5)
def test_async_op_expr():
    dataset = api.save([{"prompt": "a", "completion": "5"}])

    train_result = async_demo.train(dataset)
    model = train_result.model()
    saved_model = api.save(model, "model")

    assert (
        str(saved_model)
        == 'get("local-artifact:///list:54643291a9d1a0f06a45/obj").train().model().save("model")'
    )



import dataclasses
import typing

import weave
from weave.legacy.weave import graph, panel, panel_util


@weave.type()
class DateRangeConfig:
    domain: weave.Node[list[str]] = dataclasses.field(
        default_factory=lambda: graph.VoidNode()
    )


@weave.type()
class DateRange(panel.Panel):
    id = "DateRange"
    config: typing.Optional[DateRangeConfig] = dataclasses.field(
        default_factory=DateRangeConfig
    )

    def __init__(
        self, input_node=graph.VoidNode(), vars=None, config=None, **options
    ) -> None:
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = DateRangeConfig()
        if "domain" in options:
            self.config.domain = options["domain"]

    @weave.op()
    def value(self) -> float:
        return weave.use(self.input_node)



import contextlib
import contextvars
import logging
import typing

import numpy as np
import pyarrow as pa

from weave.legacy.weave import weave_internal
from weave.legacy.weave import errors,weavify, registry_mem
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, use
from weave.legacy.weave import dispatch, graph, graph_debug, op_args, op_def
from weave.legacy.weave.arrow import convert
from weave.legacy.weave.arrow.arrow import ArrowWeaveListType
from weave.legacy.weave.arrow.arrow_tags import pushdown_list_tags
from weave.legacy.weave.arrow.list_ import ArrowWeaveList
from weave.legacy.weave.ops_arrow import arraylist_ops, convert_ops
from weave.legacy.weave.ops_arrow.dict import preprocess_merge
from weave.legacy.weave.ops_primitives.dict import dict_


class VectorizeError(errors.WeaveBaseError):
    pass


def make_vectorized_object_constructor(constructor_op_name: str) -> None:
    constructor_op = registry_mem.memory_registry.get_op(constructor_op_name)
    if callable(constructor_op.raw_output_type):
        raise errors.WeaveInternalError(
            "Unexpected. All object type constructors have fixed return types."
        )

    type_name = constructor_op.raw_output_type.name
    vectorized_constructor_op_name = f'ArrowWeaveList-{type_name.replace("-", "_")}'
    if registry_mem.memory_registry.have_op(vectorized_constructor_op_name):
        return

    output_type = ArrowWeaveListType(constructor_op.raw_output_type)

    @op(
        name=vectorized_constructor_op_name,
        input_type={
            "attributes": ArrowWeaveListType(
                constructor_op.input_type.weave_type().property_types["attributes"]  # type: ignore
            )
        },
        output_type=output_type,
        render_info={"type": "function"},
    )
    def vectorized_constructor(attributes):
        if callable(output_type):
            ot = output_type({"attributes": types.TypeRegistry.type_of(attributes)})
        else:
            ot = output_type
        return ArrowWeaveList(
            attributes._arrow_data, ot.object_type, attributes._artifact
        )


def _create_manually_mapped_op(
    op_name: str,
    inputs: typing.Dict[str, graph.Node],
    vectorized_keys: set[str],
):
    if len(vectorized_keys) == 1:
        return _create_manually_mapped_op_singular(
            op_name, inputs, list(vectorized_keys)[0]
        )
    op = registry_mem.memory_registry.get_op(op_name)
    inputs = _vectorized_inputs_as_awl_non_vectorized_as_lists(inputs, vectorized_keys)

    mapped_inputs = {k: v for k, v in inputs.items() if k in vectorized_keys}
    rest_inputs = {k: v for k, v in inputs.items() if k not in vectorized_keys}

    from weave.legacy.weave.ops_arrow import dict

    input_arr = dict.arrow_dict_(**mapped_inputs).to_py()

    map_op = registry_mem.memory_registry.get_op("map")
    return map_op(
        input_arr,
        lambda input_dict: op(
            **{k: input_dict[k] for k in mapped_inputs}, **rest_inputs
        ),
    )


def _create_manually_mapped_op_singular(
    op_name: str,
    inputs: typing.Dict[str, graph.Node],
    vectorized_key: str,
):
    op = registry_mem.memory_registry.get_op(op_name)

    # We want to keep our vectorized inputs as AWL, even though we're going to use
    # listmap. Converting to python converts recursively, so if there are inner AWLs
    # -- as is the case in the output of AWL.groupby() which is
    # AWL<Tagged<..., AWL<...>> -- we want to keep those as AWLs. This way
    # AWL.groupby().map(lambda x: x.groupby()) will correctly keep all of the data
    # in arrow and use the arrow groupby for the inner call.

    inputs = _vectorized_inputs_as_awl_non_vectorized_as_lists(
        inputs, set([vectorized_key])
    )

    rest_inputs = {k: v for k, v in inputs.items() if k != vectorized_key}

    input_arr = inputs[vectorized_key]

    map_op = registry_mem.memory_registry.get_op("_listmap")
    return map_op(
        input_arr,
        lambda row: op(**{vectorized_key: row}, **rest_inputs),
    )


def _type_is_assignable_to_awl_list(t: types.Type) -> bool:
    return ArrowWeaveListType().assign_type(t)


def _type_is_assignable_to_py_list_not_awl_list(t: types.Type) -> bool:
    return types.List().assign_type(t) and not _type_is_assignable_to_awl_list(t)


def _ensure_list_like_node_is_awl(node: graph.Node) -> graph.Node:
    """
    Ensures that the node is an ArrowWeaveList by appending a conversion op (or stripping
    off an existing conversion op if possible)
    """
    if _type_is_assignable_to_awl_list(node.type):
        return node
    elif _type_is_assignable_to_py_list_not_awl_list(node.type):
        if (
            isinstance(node, graph.OutputNode)
            and node.from_op.name == "ArrowWeaveList-to_py"
        ):
            return list(node.from_op.inputs.values())[0]
        else:
            return convert_ops.list_to_arrow(node)
    else:
        return node


def _ensure_list_like_node_is_list(node: graph.Node) -> graph.Node:
    """
    Ensures that the node is an list by appending a conversion op (or stripping
    off an existing conversion op if possible)
    """
    if _type_is_assignable_to_py_list_not_awl_list(node.type):
        return node
    elif _type_is_assignable_to_awl_list(node.type):
        if (
            isinstance(node, graph.OutputNode)
            and node.from_op.name == "op-list_to_arrow"
        ):
            return list(node.from_op.inputs.values())[0]
        else:
            return convert_ops.to_py(node)
    else:
        return node


def _process_vectorized_inputs(
    inputs: dict[str, graph.Node],
    vectorized_keys: set[str],
    on_path: typing.Optional[typing.Callable] = None,
    off_path: typing.Optional[typing.Callable] = None,
) -> dict[str, graph.Node]:
    def identity(x):
        return x

    if on_path is None:
        on_path = identity
    if off_path is None:
        off_path = identity
    return {
        k: (on_path(in_node) if k in vectorized_keys else off_path(in_node))
        for k, in_node in inputs.items()
    }


def _vectorized_inputs_as_list(
    inputs: dict[str, graph.Node], vectorized_keys: set[str]
) -> dict[str, graph.Node]:
    return _process_vectorized_inputs(
        inputs, vectorized_keys, on_path=_ensure_list_like_node_is_list
    )


def _vectorized_inputs_as_awl(
    inputs: dict[str, graph.Node], vectorized_keys: set[str]
) -> dict[str, graph.Node]:
    return _process_vectorized_inputs(
        inputs, vectorized_keys, on_path=_ensure_list_like_node_is_awl
    )


def _vectorized_inputs_as_awl_non_vectorized_as_lists(
    inputs: dict[str, graph.Node], vectorized_keys: set[str]
) -> dict[str, graph.Node]:
    return _process_vectorized_inputs(
        inputs,
        vectorized_keys,
        on_path=_ensure_list_like_node_is_awl,
        off_path=_ensure_list_like_node_is_list,
    )


def _vectorize_lambda_output_node(node: graph.OutputNode, vectorized_keys: set[str]):
    # In a situation where we are trying to vectorize a "lambda"
    # function and the input is a a weave arrow list, then we are ina
    # bit of a pickle. This means we are trying to vectorize applying
    # this lambda to each element of the AWL. For example:
    # awl([[{"a":1, "b": 1}, {"a": 1, "b": 2}], [{"a": 2, "b": 3}, {"a": 2, "b": 4}]]).map(lambda row: row.groupby(lambda row: row["a"]))
    # When we hit the inner groupby, we are in this case. This is not
    # possible to vectorize grouping inside of a map. I think we could
    # figure out how to support nested mapping, but all the other pairs
    # are not possible to vectorize (to my knowledge). Therefore, in
    # these cases, we want to forcibly bail out to the list map which
    # does a `execute_fast.fast_map_fn` on each element of the list.
    return _create_manually_mapped_op_singular(
        node.from_op.name,
        node.from_op.inputs,
        next(iter(node.from_op.inputs)),
    )


def _is_lambda_output_node(node: graph.OutputNode):
    return (
        node.from_op.name.endswith("map")
        or node.from_op.name.endswith("groupby")
        or node.from_op.name.endswith("filter")
        or node.from_op.name.endswith("sort")
    )


def _is_non_simd_node(node: graph.OutputNode, vectorized_keys: set[str]):
    # These are ops (List/AWL) that are NOT SIMD (Single instruction, multiple data). This list is a hand
    # curated list from looking at Weave0. We probably need to refactor this entire vectorize to have a more
    # rigorous ruleset that can be applied, but in the interest of time, we are hand-crafting this for now
    non_vectorized_awl_op_names = [
        "limit",
        "offset",
        "unnest",
        "flatten",
        "2DProjection",
        "count",  # this is now SIMD
        "joinToStr",
        "unique",
        "numbers-sum",  # this is now SIMD
        "numbers-avg",  # this is now SIMD
        "numbers-argmax",
        "numbers-argmin",
        "numbers-stddev",
        "numbers-min",  # this is now SIMD
        "numbers-max",  # this is now SIMD
    ]
    first_arg_is_vectorized = list(node.from_op.inputs.keys())[0] in vectorized_keys
    return first_arg_is_vectorized and any(
        node.from_op.name.endswith(op_name) for op_name in non_vectorized_awl_op_names
    )


def _safe_get_op_for_inputs(
    name: str, inputs: dict[str, graph.Node]
) -> typing.Optional[op_def.OpDef]:
    try:
        return dispatch.get_op_for_inputs(name, {k: v.type for k, v in inputs.items()})
    except errors.WeaveDispatchError:
        return None


def _safe_get_weavified_op(op: op_def.OpDef) -> typing.Optional[graph.Node]:
    if op.weave_fn is None:
        try:
            op.weave_fn = weavify.op_to_weave_fn(op)
        except (
            errors.WeaveInternalError,
            errors.WeavifyError,
            errors.WeaveDispatchError,
            errors.WeaveTypeError,
        ):
            pass

    return op.weave_fn


def _vectorize_list_special_case(node_name, node_inputs, vectorized_keys):
    # Unfortunately, we need to check to see if the types are all the same
    # else arrow cannot make the list.
    possible_inputs = _vectorized_inputs_as_awl_non_vectorized_as_lists(
        node_inputs, vectorized_keys
    )
    running_type = None
    is_valid = True
    for v in possible_inputs.values():
        if isinstance(v.type, ArrowWeaveListType):
            obj_type = v.type.object_type
            if running_type is None:
                running_type = obj_type
            elif not running_type.assign_type(obj_type):
                is_valid = False
                break
    if is_valid:
        # TODO: If the AWL types are not all the same, it will bust here.
        op = registry_mem.memory_registry.get_op("ArrowWeaveList-vectorizedList")
        return op.lazy_call(**possible_inputs)
    else:
        return _create_manually_mapped_op(
            node_name,
            possible_inputs,
            vectorized_keys,
        )


def _vectorize_merge_special_case(node_name, node_inputs, vectorized_keys):
    self = node_inputs["lhs"]
    other = node_inputs["rhs"]

    self_ensured = preprocess_merge(self, other)
    other_ensured = preprocess_merge(other, self)

    return self_ensured.merge(other_ensured)


def vectorize(
    weave_fn,
    with_respect_to: typing.Optional[typing.Iterable[graph.VarNode]] = None,
    stack_depth: int = 0,
    strict=False,
):
    """Convert a Weave Function of T to a Weave Function of ArrowWeaveList[T]

    We walk the DAG represented by weave_fn, starting from its roots. Replace
    with_respect_to VarNodes of Type T with ArrowWeaveList[T]. Then as we
    walk up the DAG, replace OutputNodes with new op calls to whatever ops
    exist that can handle the changed input types.
    """

    # TODO: handle with_respect_to, it doesn't do anything right now.

    if stack_depth > 10:
        raise VectorizeError("Vectorize recursion depth exceeded")

    def ensure_object_constructors_created(node: graph.Node) -> graph.Node:
        if isinstance(node, graph.OutputNode):
            if node.from_op.name.startswith("objectConstructor-"):
                make_vectorized_object_constructor(node.from_op.name)
        return node

    def expand_nodes(node: graph.Node) -> graph.Node:
        if isinstance(node, graph.OutputNode):
            inputs = node.from_op.inputs
            if node.from_op.name == "number-bin":
                bin_fn = weave_internal.use(inputs["bin_fn"])
                in_ = inputs["in_"]
                return weave_internal.call_fn(bin_fn, {"row": in_})  # type: ignore
            if node.from_op.name == "timestamp-bin":
                in_ = inputs["in_"]
                res = weave_internal.better_call_fn(inputs["bin_fn"].val, in_)  # type: ignore
                return res
        return node

    def vectorize_output_node(node: graph.OutputNode, vectorized_keys: set[str]):
        # In this function, we will "vectorize" an output_node. This is a
        # shallow mapper against an output node with var nodes in its ancestry.
        # All VarNodes in the original graph will be converted to a AWL of the
        # same type. See `vectorize_along_wrt_paths` for the outer loop that
        # calls this function and performs this variable replacement.
        #
        # Moreover, we are provided a list of keys for the inputs which are in
        # the "vectorization" path. This is important bookkeeping to understand
        # if a list-like input is to be treated as a single list for each "loop"
        # of the vectorization pass (non vectorized path); or if the list-like input
        # is in the vectorization path, meaning each "loop" of the pass should "iterate"
        # over the elements. For example, consider ArrowWeaveList-vectorizedDict:
        #
        # vectorizedDict({
        #   "a": [1, 2, 3], # vectorized path
        #   "b": [4, 5, 6], # vectorized path
        # }) = [
        #   {"a": 1, "b": 4},
        #   {"a": 2, "b": 5},
        #   {"a": 3, "b": 6},
        # ]
        #
        # vectorizedDict({
        #   "a": [1, 2, 3], # vectorized path
        #   "b": [4, 5, 6], # non-vector path
        # }) = [
        #   {"a": 1, "b": [4,5,6]},
        #   {"a": 2, "b": [4,5,6]},
        #   {"a": 3, "b": [4,5,6]},
        # ]
        #
        #
        # So, the main purpose of this function is to say: given these new inputs,
        # dispatch to the correct op such that the result is properly vectorized.

        node_inputs = node.from_op.inputs
        node_name = node.from_op.name

        # First, we need to handle a a few special cases:
        # 1. If the node is a lambda function, then we know we can't vectorize it
        if _is_lambda_output_node(node):
            # Example: [[1,2,3], [3,4,5]].map(row => row.map(x => x + 1))
            return _vectorize_lambda_output_node(node, vectorized_keys)

        # 1b. If custom op, never try to vectorize for now.
        node_op_def = registry_mem.memory_registry.get_op(node.from_op.name)
        if node_op_def.location is not None:
            return _create_manually_mapped_op(node_name, node_inputs, vectorized_keys)

        # 2. If the op is `dict` or `list` then we manually hard code the vectorized version
        # since dispatch will choose the non-vectorized version. Note that we transform the inputs
        # appropriately. See comments in header of function
        if node_name == "dict":
            op = registry_mem.memory_registry.get_op("ArrowWeaveList-vectorizedDict")
            return op.lazy_call(
                **_vectorized_inputs_as_awl_non_vectorized_as_lists(
                    node_inputs, vectorized_keys
                )
            )
        if node_name == "list":
            return _vectorize_list_special_case(node_name, node_inputs, vectorized_keys)

        if node_name == "merge":
            return _vectorize_merge_special_case(
                node_name, node_inputs, vectorized_keys
            )

        # 3. In the case of `Object-__getattr__`, we need to special case it will only work when the first arg is AWL
        # and the second is a string:
        if node_name == "Object-__getattr__":
            arg_names = list(node_inputs.keys())
            if arg_names[0] in vectorized_keys and arg_names[1] not in vectorized_keys:
                op = registry_mem.memory_registry.get_op(
                    "ArrowWeaveListObject-__vectorizedGetattr__"
                )
                return op.lazy_call(
                    **{
                        arg_names[0]: _ensure_list_like_node_is_awl(
                            node_inputs[arg_names[0]]
                        ),
                        arg_names[1]: node_inputs[arg_names[1]],
                    }
                )

        input0_name = list(node_inputs.keys())[0]
        input0 = list(node_inputs.values())[0]
        if input0_name in vectorized_keys and types.List(
            types.optional(types.List())
        ).assign_type(input0.type):
            if node_name.endswith("index") or node_name.endswith("__getitem__"):
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return arraylist_ops.listindex(*inputs_as_awl.values())
            elif node_name.endswith("count"):
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return arraylist_ops.list_numbers_count(*inputs_as_awl.values())
            elif node_name.endswith("max"):
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return arraylist_ops.list_numbers_max(*inputs_as_awl.values())
            elif node_name.endswith("min"):
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return arraylist_ops.list_numbers_min(*inputs_as_awl.values())
            elif node_name.endswith("avg"):
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return arraylist_ops.list_numbers_avg(*inputs_as_awl.values())
            elif node_name.endswith("sum"):
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return arraylist_ops.list_numbers_sum(*inputs_as_awl.values())

        if node_name == "dropna":
            arg_names = list(node_inputs.keys())
            if arg_names[0] in vectorized_keys:
                op = registry_mem.memory_registry.get_op(
                    "ArrowWeaveListList-vectorizedDropna"
                )
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return op.lazy_call(*inputs_as_awl.values())
        elif node_name.endswith("isNone"):
            arg_names = list(node_inputs.keys())
            if arg_names[0] in vectorized_keys:
                op = registry_mem.memory_registry.get_op(
                    "ArrowWeaveList-vectorizedIsNone"
                )
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                return op.lazy_call(*inputs_as_awl.values())
        elif node_name.endswith("-equal") or node_name.endswith("-notEqual"):
            # if the first arg is not in the vectorized path, but is a constant, and the second is in the vectorized path,
            # then we swap the order of the inputs so that we can use existing ops we have written where the AWL is first
            # in order for tag flow to proceed as expected
            arg_names = list(node_inputs.keys())
            if arg_names[1] in vectorized_keys and arg_names[0] not in vectorized_keys:
                inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
                inputs_as_awl = {
                    arg_names[0]: inputs_as_awl[arg_names[1]],
                    arg_names[1]: inputs_as_awl[arg_names[0]],
                }
                maybe_op = _safe_get_op_for_inputs(node_name, inputs_as_awl)
                if maybe_op is not None:
                    return maybe_op.lazy_call(*inputs_as_awl.values())

        # 4. Non SIMD ops (List/AWL)
        if _is_non_simd_node(node, vectorized_keys):
            return _create_manually_mapped_op(
                node.from_op.name,
                node_inputs,
                vectorized_keys,
            )

        # Now, if we have not returned by now, then we can move on to the main logic of this function.

        # Part 1: Attempt to dispatch using the AWL inputs (This would be the most ideal case - pure AWL)
        inputs_as_awl = _vectorized_inputs_as_awl(node_inputs, vectorized_keys)
        maybe_op = _safe_get_op_for_inputs(node_name, inputs_as_awl)
        if maybe_op is not None:
            return maybe_op.lazy_call(*inputs_as_awl.values())

        # Part 2: We still want to use Arrow if possible. Here we are going to attempt to
        # weavify, then vectorize the op implementation.
        node_op = registry_mem.memory_registry.get_op(node_name)
        maybe_weavified_op = _safe_get_weavified_op(node_op)
        if maybe_weavified_op is not None:
            with_respect_to = None  # TODO: only vectorize the vectorization path!
            vectorized = vectorize(
                maybe_weavified_op, with_respect_to, stack_depth=stack_depth + 1
            )
            return weave_internal.call_fn(vectorized, inputs_as_awl)

        # Part 3: Attempt to dispatch using the list-like inputs (this is preferred to the final case)
        inputs_as_list = _vectorized_inputs_as_list(node_inputs, vectorized_keys)
        maybe_op = _safe_get_op_for_inputs(node_name, inputs_as_list)
        if maybe_op is not None and maybe_op.derived_from is None:
            return maybe_op.lazy_call(*inputs_as_list.values())

        # Mapped ops can be handled with map_each
        if node_op.derived_from is not None:
            # TODO: other arguments may also be vectorized. In that case this will crash
            input_vals = list(node_inputs.values())
            map_each_op = registry_mem.memory_registry.get_op("ArrowWeaveList-mapEach")
            return map_each_op(
                input_vals[0], lambda x: node_op.derived_from(x, *input_vals[1:])
            )
        # If an op is already an arrow version, we may still need to use mapeach
        # to lift it to another dimension. This needs to be generalized I think.
        # Added for now for timeseries plot.
        if node_op.name == "ArrowWeaveListTypedDict-pick":
            input_vals = list(node_inputs.values())
            map_each_op = registry_mem.memory_registry.get_op("ArrowWeaveList-mapEach")
            pick_op = registry_mem.memory_registry.get_op("typedDict-pick")
            return map_each_op(input_vals[0], lambda x: pick_op(x, *input_vals[1:]))

        if strict:
            raise errors.WeaveVectorizationError(f"Could not vectorize op {node_name}")
        else:
            # Final Fallback: We have no choice anymore. We must bail out completely to mapping
            # over all the vectorized inputs and calling the function directly.
            # If we hit this, then it means our vectorization has
            # created inputs which have no matching op. For example,
            # if we are doing a pick operation and the key is a
            # vectorized VarNode. This happens when picking a run
            # color using a vectorized list of runs for a table
            # (since pick(dict, list<string>) is not implemented).
            # This can happen for other ops like `add` and `mul` as
            # well (imagine `row => 1 + row`)
            #
            # In order to safely handle this case, we need to simply map
            # the original op over all the vectorized inputs.
            res = _create_manually_mapped_op(
                node_name,
                node_inputs,
                vectorized_keys,
            )
            message = (
                f"Encountered non-dispatchable op ({node_name}) during vectorization."
            )
            message += "This is likely due to vectorization path of the function not leading to the"
            message += "first parameter. Bailing out to manual mapping"
            logging.warning(message)
            return res

    # Vectorize is "with respect to" (wrt) specific variable nodes in the graph.
    # vectorize_along_wrt_paths keeps track of nodes that have already
    # been vectorized, ie nodes that have a wrt variable in their ancestry.
    # We don't try to vectorize paths for which that is not the case.
    already_vectorized_nodes: set[graph.Node] = set()

    def vectorize_along_wrt_paths(node: graph.Node):
        if isinstance(node, graph.OutputNode):
            vectorized_keys = set()
            for input_key, input_node in node.from_op.inputs.items():
                if input_node in already_vectorized_nodes:
                    vectorized_keys.add(input_key)
            if len(vectorized_keys) == 0:
                # not along vectorize path
                return node
            new_node = vectorize_output_node(node, vectorized_keys)
            already_vectorized_nodes.add(new_node)
            return new_node
        elif isinstance(node, graph.VarNode):
            # Vectorize variable
            if with_respect_to is None or any(
                node is wrt_node for wrt_node in with_respect_to
            ):
                new_node = graph.VarNode(ArrowWeaveListType(node.type), node.name)
                already_vectorized_nodes.add(new_node)
                return new_node
            # not along vectorize path
            return node
        elif isinstance(node, graph.ConstNode):
            # not along vectorize path
            return node
        else:
            raise errors.WeaveInternalError("Unexpected node: %s" % node)

    weave_fn = graph.map_nodes_top_level(
        [weave_fn], ensure_object_constructors_created
    )[0]
    weave_fn = graph.map_nodes_top_level([weave_fn], expand_nodes)[0]
    return graph.map_nodes_top_level([weave_fn], vectorize_along_wrt_paths)[0]


_raise_on_python_bailout: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "_raise_on_python_bailout", default=False
)


def _should_raise_on_python_bailout() -> bool:
    return _raise_on_python_bailout.get()


@contextlib.contextmanager
def raise_on_python_bailout():
    token = _raise_on_python_bailout.set(True)
    try:
        yield
    finally:
        _raise_on_python_bailout.reset(token)


def _call_and_ensure_awl(
    awl: ArrowWeaveList, called: graph.OutputNode
) -> ArrowWeaveList:
    from weave.legacy.weave import compile

    with compile.disable_compile():
        res = use(called)
    # Since it is possible that the result of `use` bails out of arrow due to a
    # mismatch in the types / op support. This is most likely due to gap in the
    # implementation of vectorized ops. However, there are cases where it is
    # currently expected - for example calling a custom op on a custom type. An
    # example of this is in `ops_arrow/test_arrow.py::test_custom_types_tagged`:
    #
    #     ` data_node.map(lambda row: row["im"].width_())`
    #
    # If such cases did not exist, then we should probably raise in this case.
    # However, for now, we will just convert the result back to arrow if it is a
    # list.
    if not isinstance(res, ArrowWeaveList):
        err_msg = f"Applying vectorized function {called} to awl of {awl.object_type} \
            resulted in a non vectorized result type: {type(res)}. This likely \
            means 1 or more ops in the function were converted to the list \
            implementation in compile."
        if isinstance(res, list):
            res = convert.to_arrow(res)
            logging.warning(err_msg)
            if _should_raise_on_python_bailout():
                raise errors.WeaveVectorizationError(err_msg)
        else:
            raise errors.WeaveVectorizationError(err_msg)

    return res


def _call_vectorized_fn_node_maybe_awl(
    awl: ArrowWeaveList, vectorized_fn: graph.OutputNode
) -> graph.OutputNode:
    index_awl: ArrowWeaveList[int] = ArrowWeaveList(pa.array(np.arange(len(awl))))
    row_type = ArrowWeaveListType(awl.object_type)
    try:
        fn_res_node = weave_internal.call_fn(
            vectorized_fn,
            {
                "row": weave_internal.make_const_node(row_type, awl),
                "index": weave_internal.make_const_node(
                    ArrowWeaveListType(types.Int()), index_awl
                ),
            },
        )
    except errors.WeaveMissingVariableError as e:
        raise errors.WeaveBadRequest('Invalid variable "%s" in function' % e.args[0])
    return typing.cast(graph.OutputNode, weave_internal.refine_graph(fn_res_node))


def _apply_fn_node_with_tag_pushdown(
    awl: ArrowWeaveList, fn: graph.OutputNode
) -> ArrowWeaveList:
    tagged_awl = pushdown_list_tags(awl)
    return _apply_fn_node(tagged_awl, fn)


def _ensure_variadic_fn(
    fn: graph.OutputNode, dummy_var_type: types.Type, dummy_var_name: str = "row"
) -> graph.OutputNode:
    # Check if fn contains variables.
    contains_vars = len(graph.expr_vars(fn)) > 0
    # if it does, return early.
    if contains_vars:
        return fn
    # else, create a new graph which contains a row
    # variable. We will use a dict followed by a pick.
    return dict_(a=fn, b=graph.VarNode(dummy_var_type, dummy_var_name))["a"]


def _apply_fn_node(awl: ArrowWeaveList, fn: graph.OutputNode) -> ArrowWeaveList:
    debug_str = graph_debug.node_expr_str_full(fn)
    logging.info("Vectorizing: %s", debug_str)

    if len(awl) == 0:
        # Short circuit empty list for performance reasons and to avoid calling
        # aggregations on empty lists.
        logging.info(
            "Short circuiting vectorization on %s because it is empty.", debug_str
        )
        return convert.to_arrow([], types.List(fn.type), artifact=awl._artifact)

    from weave.legacy.weave import execute_fast

    fn = execute_fast._resolve_static_branches(fn)
    logging.info("Vectorizing. Static branch resolution complete.: %s", debug_str)

    vecced = vectorize(_ensure_variadic_fn(fn, awl.object_type))
    debug_str = graph_debug.node_expr_str_full(vecced)
    logging.info("Vectorizing. Vectorized: %s", debug_str)
    called = _call_vectorized_fn_node_maybe_awl(awl, vecced)
    return _call_and_ensure_awl(awl, called)



import copy
import datetime
from collections import OrderedDict, defaultdict
from typing import Any, Dict, Optional, cast

from weave.trace_server import refs_internal as ri
from weave.trace_server import trace_server_interface as tsi


def make_feedback_query_req(
    project_id: str,
    calls: list[dict[str, Any]],
) -> tsi.FeedbackQueryReq:
    # make list of weave refs to calls, to be used in feedback query
    call_refs = []
    for call in calls:
        ref = ri.InternalCallRef(project_id=call["project_id"], id=call["id"])
        call_refs.append(ref.uri())

    # construct mogo style query
    query = tsi.Query(
        **{
            "$expr": {
                "$in": [
                    {"$getField": "weave_ref"},
                    [{"$literal": call_ref} for call_ref in call_refs],
                ]
            }
        }
    )
    feedback_query_req = tsi.FeedbackQueryReq(
        project_id=project_id,
        fields=[
            "feedback_type",
            "weave_ref",
            "payload",
            "creator",
            "created_at",
            "wb_user_id",
        ],
        query=query,
    )
    return feedback_query_req


def hydrate_calls_with_feedback(
    calls: list[dict[str, Any]], feedback: tsi.FeedbackQueryRes
) -> None:
    """Hydrate calls with feedback inplace."""
    feedback_map = defaultdict(list)
    # map feedback to calls
    for feedback_item in feedback.result:
        uri = ri.parse_internal_uri(feedback_item["weave_ref"])
        if isinstance(uri, ri.InternalCallRef):
            feedback_map[uri.id].append(feedback_item)

    for call in calls:
        feedback_items = feedback_map.get(call["id"], [])
        if not call.get("summary"):
            call["summary"] = {}
        if not call["summary"].get("weave"):
            call["summary"]["weave"] = {}
        call["summary"]["weave"]["feedback"] = feedback_items


def make_derived_summary_fields(
    summary: Dict[str, Any],
    op_name: str,
    started_at: Optional[datetime.datetime] = None,
    ended_at: Optional[datetime.datetime] = None,
    exception: Optional[str] = None,
    display_name: Optional[str] = None,
) -> tsi.SummaryMap:
    """
    Make derived summary fields for a call.

    Summary is controlled by the user, but the `weave` summary key is
    used to store derived fields, adhering to the tsi.SummaryMap type.
    """
    weave_summary = summary.pop("weave", {})

    status = tsi.TraceStatus.SUCCESS
    if exception:
        status = tsi.TraceStatus.ERROR
    elif ended_at is None:
        status = tsi.TraceStatus.RUNNING
    weave_summary["status"] = status

    if ended_at and started_at:
        days = (ended_at - started_at).days
        seconds = (ended_at - started_at).seconds
        milliseconds = (ended_at - started_at).microseconds // 1000
        weave_summary["latency_ms"] = (
            days * 24 * 60 * 60 + seconds
        ) * 1000 + milliseconds

    if display_name:
        weave_summary["display_name"] = display_name
    else:
        if ri.string_will_be_interpreted_as_ref(op_name):
            op = ri.parse_internal_uri(op_name)
            if isinstance(op, ri.InternalObjectRef):
                weave_summary["trace_name"] = op.name
        else:
            weave_summary["trace_name"] = op_name

    summary["weave"] = weave_summary
    return cast(tsi.SummaryMap, summary)


def empty_str_to_none(val: Optional[str]) -> Optional[str]:
    return val if val != "" else None


def get_nested_key(d: Dict[str, Any], col: str) -> Optional[Any]:
    """
    Get a nested key from a dict. None if not found.

    Example:
    get_nested_key({"a": {"b": {"c": "d"}}}, "a.b.c") -> "d"
    get_nested_key({"a": {"b": {"c": "d"}}}, "a.b") -> {"c": "d"}
    get_nested_key({"a": {"b": {"c": "d"}}}, "foobar") -> None
    """

    def _get(data: Optional[Any], key: str) -> Optional[Any]:
        if not data or not isinstance(data, dict):
            return None
        return data.get(key)

    keys = col.split(".")
    curr: Optional[Any] = d
    for key in keys[:-1]:
        curr = _get(curr, key)
    return _get(curr, keys[-1])


def set_nested_key(d: Dict[str, Any], col: str, val: Any) -> None:
    """
    Set a nested key in a dict.

    Example:
    set_nested_key({"a": {"b": "c"}}, "a.b", "e") -> {"a": {"b": "e"}}
    set_nested_key({"a": {"b": "e"}}, "a.b.c", "e") -> {"a": {"b": {"c": "e"}}}
    """
    keys = col.split(".")
    if not keys[-1]:
        # If the columns is misformatted just return (ex: "a.b.")
        return

    curr = d
    for key in keys[:-1]:
        if key not in curr or not isinstance(curr[key], dict):
            curr[key] = {}
        curr = curr[key]
    curr[keys[-1]] = copy.deepcopy(val)


class LRUCache(OrderedDict):
    def __init__(self, max_size: int = 1000, *args: Any, **kwargs: Dict[str, Any]):
        self.max_size = max_size
        super().__init__(*args, **kwargs)

    def __setitem__(self, key: str, value: Any) -> None:
        if key not in self and len(self) >= self.max_size:
            self.popitem(last=False)
        super().__setitem__(key, value)



import weave
from weave.legacy.weave import panel


@weave.type(__override_name="wb_trace_tree-traceViewer")  # type: ignore
class PanelWBTraceTreeTraceViewer(panel.Panel):
    id = "wb_trace_tree-traceViewer"


@weave.type(__override_name="wb_trace_tree-modelViewer")  # type: ignore
class PanelWBTraceTreeModelViewer(panel.Panel):
    id = "wb_trace_tree-modelViewer"



import re
import typing
import urllib

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op
from weave.legacy.weave.gql_op_plugin import wb_gql_op_plugin
from weave.legacy.weave.ops_domain import wb_domain_types as wdt
from weave.legacy.weave.ops_domain.wandb_domain_gql import (
    gql_direct_edge_op,
    gql_prop_op,
)

# Section 1/6: Tag Getters
#


# Section 2/6: Root Ops
@op(
    name="root-allReportsGQLResolver",
    input_type={"gql_result": types.TypedDict({})},
    output_type=types.List(wdt.ReportType),
    hidden=True,
)
def root_all_reports_gql_resolver(gql_result):
    return [
        wdt.Report.from_keys(report["node"])
        for report in gql_result["instance"]["views_500"]["edges"]
        if report["node"]["type"] == "runs"  # yes, this is how we filter to Reports!?
    ]


@op(
    name="root-allReports",
    input_type={},
    output_type=types.List(wdt.ReportType),
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
        instance {{
            views_500: views(limit: 500) {{
                edges {{
                    node {{
                        type
                        {wdt.Report.REQUIRED_FRAGMENT}
                        {inner}
                    }}
                }}
            }}
        }}""",
        is_root=True,
        root_resolver=root_all_reports_gql_resolver,
    ),
)
def root_all_reports():
    raise errors.WeaveGQLCompileError(
        "root-allReports should not be executed directly. If you see this error, it is a bug in the Weave compiler."
    )


# Section 3/6: Attribute Getters
gql_prop_op("report-internalId", wdt.ReportType, "id", types.String(), True)
gql_prop_op(
    "report-name",
    wdt.ReportType,
    "displayName",
    types.String(),
)
gql_prop_op(
    "report-internalId",
    wdt.ReportType,
    "id",
    types.String(),
)
gql_prop_op(
    "report-name",
    wdt.ReportType,
    "displayName",
    types.String(),
)
gql_prop_op(
    "report-createdAt",
    wdt.ReportType,
    "createdAt",
    types.Timestamp(),
)
gql_prop_op(
    "report-updatedAt",
    wdt.ReportType,
    "updatedAt",
    types.Timestamp(),
)
gql_prop_op(
    "report-viewcount",
    wdt.ReportType,
    "viewCount",
    types.Number(),
)

# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "report-project",
    wdt.ReportType,
    "project",
    wdt.ProjectType,
)

gql_direct_edge_op(
    "report-creator",
    wdt.ReportType,
    "user",
    wdt.UserType,
)

# Section 5/6: Connection Ops
#

# Section 6/6: Non Standard Business Logic Ops


# Logic taken exactly from Weave0 / frontend
def make_name_and_id(id: str, name: typing.Optional[str]) -> str:
    id = id.replace("=", "")
    if name is None:
        return id
    name = re.sub(r"\W", "-", name)
    name = re.sub(r"-+", "-", name)
    name = urllib.parse.quote(name)

    return f"{name}--{id}"


@op(
    name="report-link",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
    id
    displayName
    project {
        id
        name 
        entity {
            id
            name
        }
    }
"""
    ),
)
def link(report: wdt.Report) -> wdt.Link:
    project = report["project"]
    entity = project["entity"]

    project_name = project["name"]
    entity_name = entity["name"]
    report_id = report["id"]
    report_name = report["displayName"]

    url = f"/{entity_name}/{project_name}/reports/{make_name_and_id(report_id, report_name)}"
    return wdt.Link(report_name, url)



import copy
import dataclasses
import typing

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave import (
    errors, 
    codifiable_value_mixin,
    codify,
    dispatch,
    graph,
    panel,
    panel_util,
)
from weave.legacy.weave.panels.bank import (
    default_panel_bank_flow_section_config,
    flow_layout,
)
from weave.legacy.weave.panels.panel_group_panel_info import PanelInfo

ItemsType = typing.TypeVar("ItemsType")

# I pulled in all the type information from PanelBank, but we're not
# using it all yet.


class LayoutParameters(typing.TypedDict):
    x: int
    y: int
    w: int
    h: int


class LayedOutPanel(typing.TypedDict):
    id: str
    layout: LayoutParameters


class PanelBankFlowSectionConfig(typing.TypedDict):
    snapToColumns: bool
    columnsPerPage: int
    rowsPerPage: int
    gutterWidth: int
    boxWidth: int
    boxHeight: int


class PanelBankSectionConfig(typing.TypedDict):
    id: str
    name: str
    panels: typing.List[LayedOutPanel]
    isOpen: bool
    flowConfig: PanelBankFlowSectionConfig
    type: str  # this 'grid' | 'flow' in ts
    sorted: int  # This is an enum in js


@weave.type()
class GroupConfig(typing.Generic[ItemsType]):
    layoutMode: str = dataclasses.field(default_factory=lambda: "vertical")
    showExpressions: typing.Union[bool, typing.Literal["editable"]] = dataclasses.field(
        default_factory=lambda: False
    )
    equalSize: bool = dataclasses.field(default_factory=lambda: False)
    style: str = dataclasses.field(default_factory=lambda: "")
    items: ItemsType = dataclasses.field(default_factory=dict)  # type: ignore
    panelInfo: typing.Optional[dict[str, typing.Any]] = dataclasses.field(
        default_factory=lambda: None
    )
    gridConfig: typing.Optional[PanelBankSectionConfig] = dataclasses.field(
        default_factory=lambda: None
    )
    liftChildVars: typing.Optional[bool] = dataclasses.field(
        default_factory=lambda: None
    )
    allowedPanels: typing.Optional[list[str]] = dataclasses.field(
        default_factory=lambda: None
    )
    enableAddPanel: typing.Optional[bool] = dataclasses.field(
        default_factory=lambda: None
    )
    disableDeletePanel: typing.Optional[bool] = dataclasses.field(
        default_factory=lambda: None
    )
    childNameBase: typing.Optional[str] = dataclasses.field(
        default_factory=lambda: None
    )


GroupConfigType = typing.TypeVar("GroupConfigType")


@dataclasses.dataclass
class GroupLayoutFlow:
    rows: int
    columns: int


@dataclasses.dataclass
class GroupPanelLayout:
    x: int
    y: int
    h: int
    w: int


@dataclasses.dataclass
class GroupPanel:
    panel: typing.Any
    id: typing.Optional[str] = None
    hidden: typing.Optional[bool] = None
    # Only used inside a grid layout
    layout: typing.Optional[GroupPanelLayout] = None


@weave.type()
class Group(panel.Panel, codifiable_value_mixin.CodifiableValueMixin):
    id = "Group"
    config: typing.Optional[GroupConfig] = dataclasses.field(
        default_factory=lambda: None
    )
    # items: typing.TypeVar("items") = dataclasses.field(default_factory=dict)

    def __init__(
        self, input_node=graph.VoidNode(), vars=None, config=None, **options
    ) -> None:
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = GroupConfig()
        if "layoutMode" in options:
            layout_mode = options["layoutMode"]
            if isinstance(layout_mode, GroupLayoutFlow):
                self.config.layoutMode = "flow"
                self.config.gridConfig = flow_layout(
                    layout_mode.rows, layout_mode.columns
                )
            else:
                self.config.layoutMode = layout_mode
                self.config.gridConfig = default_panel_bank_flow_section_config()
        if "items" in options:
            if isinstance(options["items"], dict):
                self.config.items = options["items"]
            else:
                options_dict = {}
                panel_info = {}
                if self.config.gridConfig is None:
                    self.config.gridConfig = default_panel_bank_flow_section_config()

                for o in options["items"]:
                    if isinstance(o, GroupPanel):
                        options_dict[o.id] = o.panel
                        panel_info[o.id] = {"hidden": o.hidden}
                        if o.id and o.layout:
                            self.config.gridConfig["panels"].append(
                                {
                                    "id": o.id,
                                    "layout": {
                                        "x": o.layout.x,
                                        "y": o.layout.y,
                                        "w": o.layout.w,
                                        "h": o.layout.h,
                                    },
                                }
                            )
                    else:
                        raise ValueError("Items must be GroupPanel")
                self.config.items = options_dict
                if panel_info:
                    self.config.panelInfo = panel_info  # type: ignore

        if "showExpressions" in options:
            self.config.showExpressions = options["showExpressions"]
        if "enableAddPanel" in options:
            self.config.enableAddPanel = options["enableAddPanel"]
        if "equalSize" in options:
            self.config.equalSize = options["equalSize"]
        if "style" in options:
            self.config.style = options["style"]
        self._normalize()

    def _normalize(self, frame=None) -> dict[str, list[typing.Union[str, int]]]:
        if frame is None:
            frame = {}
        frame = copy.copy(frame)
        frame.update(self.vars)
        frame["input"] = self.input_node

        items = {}
        all_missing_vars = {}
        config = typing.cast(GroupConfig, self.config)
        for name, p in config.items.items():
            injected, missing_vars = panel.run_variable_lambdas(p, frame)
            if missing_vars:
                child = injected
            else:
                child = panel_util.child_item(injected)
                child_missing_vars = None
                if isinstance(child, panel.Panel):
                    child_missing_vars = child._normalize(frame)
                    if child_missing_vars:
                        for k, v in child_missing_vars.items():
                            child_path = [name] + v
                            missing_vars[k] = child_path
                if isinstance(child, Group):
                    # lift group vars
                    child_config = typing.cast(GroupConfig, child.config)
                    for child_name, child_item in child_config.items.items():
                        if child_missing_vars:
                            continue
                        if not isinstance(child_item, graph.Node) and not isinstance(
                            child_item, panel.Panel
                        ):
                            raise ValueError(
                                "Group items must be panel.Panel or graph.Node, but panel at key '%s' is %s"
                                % (child_name, type(child_item))
                            )
                        frame[child_name] = weave_internal.make_var_node(
                            weave.type_of(child_item), child_name
                        )
            items[name] = child

            # We build up config one item at a time. Construct a version
            # with the current items so that we can do type_of on it (type_of
            # will fail if we have a lambda in the config).
            all_missing_vars.update(missing_vars)
            if not all_missing_vars:
                partial_config = dataclasses.replace(self.config, items=items)
                config_var = weave_internal.make_var_for_value(partial_config, "self")
                frame[name] = config_var.items[name]  # type: ignore

        config.items = items
        return all_missing_vars

    def add(
        self,
        name: str,
        node_or_panel: typing.Any,
        hidden: bool = False,
        layout: typing.Optional[GroupPanelLayout] = None,
    ) -> dispatch.RuntimeVarNode:
        config = typing.cast(GroupConfig, self.config)
        config.items[name] = node_or_panel
        if config.panelInfo is None:
            config.panelInfo = {}
        config.panelInfo[name] = {"hidden": hidden}
        if layout is not None:
            if config.gridConfig is None:
                config.gridConfig = default_panel_bank_flow_section_config()
            config.gridConfig["panels"].append(
                {
                    "id": name,
                    "layout": {
                        "x": layout.x,
                        "y": layout.y,
                        "w": layout.w,
                        "h": layout.h,
                    },
                }
            )
        missing_vars = self._normalize()
        if missing_vars:
            raise errors.WeaveApiError(
                "References to the following variables were not resolved at paths: %s"
                % missing_vars
            )
        return weave_internal.make_var_for_value(node_or_panel, name)  # type: ignore

    def finalize(self):
        missing_vars = self._normalize()
        if missing_vars:
            raise errors.WeaveApiError(
                "References to the following variables were not resolved at paths: %s"
                % missing_vars
            )

    def to_code(self) -> typing.Optional[str]:
        field_vals: list[tuple[str, str]] = []

        if self.config and self.config:
            gc = self.config

            if gc.equalSize != False:
                return None
            if gc.style != "":
                return None
            if (
                gc.gridConfig is not None
                and gc.gridConfig != default_panel_bank_flow_section_config()
            ):
                return None
            if gc.liftChildVars is not None:
                return None
            if gc.allowedPanels is not None:
                return None
            if gc.childNameBase is not None:
                return None

            if self.config.layoutMode != "vertical":
                field_vals.append(("layoutMode", codify.object_to_code(gc.layoutMode)))
            if gc.showExpressions != False:
                field_vals.append(
                    ("showExpressions", codify.object_to_code(gc.showExpressions))
                )
            if hasattr(gc, "layered") and gc.layered != False:
                field_vals.append(("layered", codify.object_to_code(gc.layered)))
            if gc.enableAddPanel != None and gc.enableAddPanel != False:
                field_vals.append(
                    ("enableAddPanel", codify.object_to_code(gc.enableAddPanel))
                )
            prior_vars: list[str] = []
            code_items_map = {}
            if gc.items != {}:
                for item_name, item in gc.items.items():
                    code_items_map[
                        item_name
                    ] = codify.lambda_wrapped_object_to_code_no_format(item, prior_vars)
                    prior_vars.append(item_name)
                if len(code_items_map) > 0:
                    items_val = (
                        ",".join(
                            ['"' + n + '":' + i for n, i in code_items_map.items()]
                        )
                        + ","
                    )
                    items_val = "{" + items_val + "}"
                    field_vals.append(("items", items_val))

        input_node_str = ""
        if not isinstance(self.input_node, graph.VoidNode):
            input_node_str = codify.object_to_code(self.input_node) + ","

        param_str = ""
        if len(field_vals) > 0:
            param_str = (
                ",".join([f_name + "=" + f_val for f_name, f_val in field_vals]) + ","
            )
        return (
            f"""weave.legacy.weave.panels.panel_group.Group({input_node_str} {param_str})"""
        )

    # @property
    # def config(self):
    #     self._normalize()
    #     return {"items": {k: i.to_json() for k, i in self.items.items()}}



from weave.legacy.weave import api as weave
from weave.legacy.weave import compile, ops
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType
from weave.legacy.weave.ops_domain import wb_domain_types as wdt
from weave.legacy.weave.ops_domain.project_ops import root_all_projects
from weave.legacy.weave.ops_domain.report_ops import root_all_reports

from .test_wb import table_mock1_no_display_name


def test_gql_compilation_with_keys():
    project_node = ops.project("stacey", "mendeleev")
    runs_node = project_node.runs()
    cell_node = runs_node.limit(1).id()
    compiled_node = compile.compile([cell_node])[0]

    run_type_keys = types.TypedDict(
        property_types={
            "id": types.String(),
            "name": types.String(),
        }
    )

    run_edges_type = types.TypedDict(
        property_types={
            "edges": types.List(
                object_type=types.TypedDict(property_types={"node": run_type_keys})
            )
        }
    )

    expected = TaggedValueType(
        types.TypedDict(
            {
                "project": wdt.ProjectType.with_keys(
                    {
                        "id": types.String(),
                        "name": types.String(),
                        "runs_c1233b7003317090ab5e2a75db4ad965": run_edges_type,
                    }
                )
            }
        ),
        types.List(
            TaggedValueType(
                types.TypedDict(
                    {"run": wdt.RunType.with_keys(run_type_keys.property_types)}
                ),
                types.String(),
            )
        ),
    )

    assert compiled_node.type == expected


def test_gql_compilation_root_op_custom_key_fn():
    root = root_all_projects().limit(1)[0].runs().id()
    compiled_node = compile.compile([root])[0]

    run_type = wdt.RunType.with_keys({"id": types.String(), "name": types.String()})
    project_type = wdt.ProjectType.with_keys(
        {
            "id": types.String(),
            "name": types.String(),
            "runs_c1233b7003317090ab5e2a75db4ad965": types.TypedDict(
                {
                    "edges": types.List(
                        types.TypedDict({"node": types.TypedDict(run_type.keys)})
                    )
                }
            ),
        }
    )

    expected = TaggedValueType(
        types.TypedDict({"project": project_type}),
        types.List(TaggedValueType(types.TypedDict({"run": run_type}), types.String())),
    )

    assert compiled_node.type == expected


def test_project_artifacts():
    node = ops.project("stacey", "mendeleev").artifacts().id()
    compiled_node = compile.compile([node])[0]
    expected = TaggedValueType(
        types.TypedDict(
            {
                "project": wdt.ProjectType.with_keys(
                    {
                        "id": types.String(),
                        "name": types.String(),
                        "artifactTypes_100": types.TypedDict(
                            {
                                "edges": types.List(
                                    types.TypedDict(
                                        {
                                            "node": types.TypedDict(
                                                {
                                                    "id": types.String(),
                                                    "artifactCollections_100": types.TypedDict(
                                                        {
                                                            "edges": types.List(
                                                                types.TypedDict(
                                                                    {
                                                                        "node": types.TypedDict(
                                                                            {
                                                                                "id": types.String(),
                                                                                "name": types.String(),
                                                                            }
                                                                        )
                                                                    }
                                                                )
                                                            )
                                                        }
                                                    ),
                                                }
                                            )
                                        }
                                    )
                                )
                            }
                        ),
                    }
                )
            }
        ),
        types.List(types.String()),
    )

    assert compiled_node.type == expected


def test_typename():
    node = ops.project("stacey", "mendeleev").artifacts()[0].isPortfolio()
    compiled_node = compile.compile([node])[0]
    expected = TaggedValueType(
        types.TypedDict(
            {
                "project": wdt.ProjectType.with_keys(
                    {
                        "id": types.String(),
                        "name": types.String(),
                        "artifactTypes_100": types.TypedDict(
                            {
                                "edges": types.List(
                                    types.TypedDict(
                                        {
                                            "node": types.TypedDict(
                                                {
                                                    "id": types.String(),
                                                    "artifactCollections_100": types.TypedDict(
                                                        {
                                                            "edges": types.List(
                                                                types.TypedDict(
                                                                    {
                                                                        "node": types.TypedDict(
                                                                            {
                                                                                "__typename": types.Const(
                                                                                    types.String(),
                                                                                    "ArtifactCollection",
                                                                                ),
                                                                                "id": types.String(),
                                                                                "name": types.String(),
                                                                            }
                                                                        )
                                                                    }
                                                                )
                                                            )
                                                        }
                                                    ),
                                                }
                                            )
                                        }
                                    )
                                )
                            }
                        ),
                    }
                )
            }
        ),
        types.Boolean(),
    )

    assert compiled_node.type == expected


def test_last_membership():
    node = ops.project("stacey", "mendeleev").artifacts()[0].lastMembership()
    compiled_node = compile.compile([node])[0]
    expected = TaggedValueType(
        types.TypedDict(
            {
                "project": wdt.ProjectType.with_keys(
                    {
                        "id": types.String(),
                        "name": types.String(),
                        "artifactTypes_100": types.TypedDict(
                            {
                                "edges": types.List(
                                    types.TypedDict(
                                        {
                                            "node": types.TypedDict(
                                                {
                                                    "id": types.String(),
                                                    "artifactCollections_100": types.TypedDict(
                                                        {
                                                            "edges": types.List(
                                                                types.TypedDict(
                                                                    {
                                                                        "node": types.TypedDict(
                                                                            {
                                                                                "id": types.String(),
                                                                                "name": types.String(),
                                                                                "artifactMemberships_first_1": types.TypedDict(
                                                                                    {
                                                                                        "edges": types.List(
                                                                                            types.TypedDict(
                                                                                                {
                                                                                                    "node": types.TypedDict(
                                                                                                        {
                                                                                                            "id": types.String(),
                                                                                                        }
                                                                                                    )
                                                                                                }
                                                                                            )
                                                                                        )
                                                                                    }
                                                                                ),
                                                                            }
                                                                        )
                                                                    }
                                                                )
                                                            )
                                                        }
                                                    ),
                                                }
                                            )
                                        }
                                    )
                                )
                            }
                        ),
                    }
                )
            }
        ),
        types.optional(
            wdt.ArtifactCollectionMembershipType.with_keys({"id": types.String()})
        ),
    )

    assert compiled_node.type == expected


def test_mapped_last_membership():
    node = ops.project("stacey", "mendeleev").artifacts().lastMembership()
    compiled_node = compile.compile([node])[0]
    expected = TaggedValueType(
        types.TypedDict(
            {
                "project": wdt.ProjectType.with_keys(
                    {
                        "id": types.String(),
                        "name": types.String(),
                        "artifactTypes_100": types.TypedDict(
                            {
                                "edges": types.List(
                                    types.TypedDict(
                                        {
                                            "node": types.TypedDict(
                                                {
                                                    "id": types.String(),
                                                    "artifactCollections_100": types.TypedDict(
                                                        {
                                                            "edges": types.List(
                                                                types.TypedDict(
                                                                    {
                                                                        "node": types.TypedDict(
                                                                            {
                                                                                "id": types.String(),
                                                                                "name": types.String(),
                                                                                "artifactMemberships_first_1": types.TypedDict(
                                                                                    {
                                                                                        "edges": types.List(
                                                                                            types.TypedDict(
                                                                                                {
                                                                                                    "node": types.TypedDict(
                                                                                                        {
                                                                                                            "id": types.String(),
                                                                                                        }
                                                                                                    )
                                                                                                }
                                                                                            )
                                                                                        )
                                                                                    }
                                                                                ),
                                                                            }
                                                                        )
                                                                    }
                                                                )
                                                            )
                                                        }
                                                    ),
                                                }
                                            )
                                        }
                                    )
                                )
                            }
                        ),
                    }
                )
            }
        ),
        types.List(
            types.optional(
                wdt.ArtifactCollectionMembershipType.with_keys({"id": types.String()})
            )
        ),
    )

    assert compiled_node.type == expected


def test_op_with_refiner(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock1_no_display_name)
    node = ops.project("stacey", "mendeleev").runs()[0].summary()["legacy_table"]
    compiled_node = compile.compile([node])[0]

    # it was hard to manually construct the type here in python, so i dumped it to
    # a dict in the debugger and re-created the type from that below :).
    expected_compiled_type_dict = {
        "type": "tagged",
        "tag": {
            "type": "typedDict",
            "propertyTypes": {
                "run": {
                    "type": "tagged",
                    "tag": {
                        "type": "typedDict",
                        "propertyTypes": {
                            "project": {
                                "type": "PartialObject",
                                "keys": {
                                    "id": "string",
                                    "name": "string",
                                    "runs_c1233b7003317090ab5e2a75db4ad965": {
                                        "type": "typedDict",
                                        "propertyTypes": {
                                            "edges": {
                                                "type": "list",
                                                "objectType": {
                                                    "type": "typedDict",
                                                    "propertyTypes": {
                                                        "node": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "id": "string",
                                                                "name": "string",
                                                                "project": {
                                                                    "type": "typedDict",
                                                                    "propertyTypes": {
                                                                        "id": "string",
                                                                        "name": "string",
                                                                        "entity": {
                                                                            "type": "typedDict",
                                                                            "propertyTypes": {
                                                                                "id": "string",
                                                                                "name": "string",
                                                                            },
                                                                        },
                                                                    },
                                                                },
                                                                "summaryMetricsSubset": "string",
                                                            },
                                                        }
                                                    },
                                                },
                                            }
                                        },
                                    },
                                },
                                "keyless_weave_type_class": "project",
                            }
                        },
                    },
                    "value": {
                        "type": "PartialObject",
                        "keys": {
                            "id": "string",
                            "name": "string",
                            "project": {
                                "type": "typedDict",
                                "propertyTypes": {
                                    "id": "string",
                                    "name": "string",
                                    "entity": {
                                        "type": "typedDict",
                                        "propertyTypes": {
                                            "id": "string",
                                            "name": "string",
                                        },
                                    },
                                },
                            },
                            "summaryMetricsSubset": "string",
                        },
                        "keyless_weave_type_class": "run",
                    },
                },
                "project": {
                    "type": "PartialObject",
                    "keys": {
                        "id": "string",
                        "name": "string",
                        "runs_c1233b7003317090ab5e2a75db4ad965": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "edges": {
                                    "type": "list",
                                    "objectType": {
                                        "type": "typedDict",
                                        "propertyTypes": {
                                            "node": {
                                                "type": "typedDict",
                                                "propertyTypes": {
                                                    "id": "string",
                                                    "name": "string",
                                                    "summaryMetricsSubset": {
                                                        "type": "union",
                                                        "members": ["none", "string"],
                                                    },
                                                    "project": {
                                                        "type": "typedDict",
                                                        "propertyTypes": {
                                                            "id": "string",
                                                            "name": "string",
                                                            "entity": {
                                                                "type": "typedDict",
                                                                "propertyTypes": {
                                                                    "id": "string",
                                                                    "name": "string",
                                                                },
                                                            },
                                                        },
                                                    },
                                                },
                                            }
                                        },
                                    },
                                }
                            },
                        },
                    },
                    "keyless_weave_type_class": "project",
                },
            },
        },
        "value": {
            "type": "file",
            "_base_type": {"type": "FileBase"},
            "extension": "json",
            "wbObjectType": {
                "type": "table",
                "_base_type": {"type": "Object"},
                "_is_object": True,
                "_rows": {
                    "type": "ArrowWeaveList",
                    "_base_type": {"type": "list"},
                    "objectType": {"type": "typedDict", "propertyTypes": {}},
                },
            },
        },
    }

    expected = types.TypeRegistry.type_from_dict(expected_compiled_type_dict)
    assert compiled_node.type == expected


def test_gql_connection_op():
    node = ops.project("stacey", "mendeleev").artifactTypes()[0].artifactVersions()
    compiled_node = compile.compile([node])[0]

    expected_compiled_type_dict = {
        "type": "tagged",
        "tag": {
            "type": "typedDict",
            "propertyTypes": {
                "project": {
                    "type": "PartialObject",
                    "keys": {
                        "id": "string",
                        "name": "string",
                        "artifactTypes": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "edges": {
                                    "type": "list",
                                    "objectType": {
                                        "type": "typedDict",
                                        "propertyTypes": {
                                            "node": {
                                                "type": "typedDict",
                                                "propertyTypes": {
                                                    "id": "string",
                                                    "name": "string",
                                                    "artifactCollections_c1233b7003317090ab5e2a75db4ad965": {
                                                        "type": "typedDict",
                                                        "propertyTypes": {
                                                            "edges": {
                                                                "type": "list",
                                                                "objectType": {
                                                                    "type": "typedDict",
                                                                    "propertyTypes": {
                                                                        "node": {
                                                                            "type": "typedDict",
                                                                            "propertyTypes": {
                                                                                "artifacts_c1233b7003317090ab5e2a75db4ad965": {
                                                                                    "type": "typedDict",
                                                                                    "propertyTypes": {
                                                                                        "edges": {
                                                                                            "type": "list",
                                                                                            "objectType": {
                                                                                                "type": "typedDict",
                                                                                                "propertyTypes": {
                                                                                                    "node": {
                                                                                                        "type": "typedDict",
                                                                                                        "propertyTypes": {
                                                                                                            "id": "string"
                                                                                                        },
                                                                                                    }
                                                                                                },
                                                                                            },
                                                                                        }
                                                                                    },
                                                                                }
                                                                            },
                                                                        }
                                                                    },
                                                                },
                                                            }
                                                        },
                                                    },
                                                },
                                            }
                                        },
                                    },
                                }
                            },
                        },
                    },
                    "keyless_weave_type_class": "project",
                }
            },
        },
        "value": {
            "type": "list",
            "objectType": {
                "type": "PartialObject",
                "keys": {"id": "string"},
                "keyless_weave_type_class": "artifactVersion",
            },
        },
    }

    expected = types.TypeRegistry.type_from_dict(expected_compiled_type_dict)

    assert compiled_node.type == expected


def test_early_termination_of_gql_key_propagation(fake_wandb):
    # Test to ensure that an operation without a defined GQL key propagation function
    # (gql_op_output_type) still executes. The result can still be serialized and deserialized,
    # even without the keys for the result.

    # Add a mock to the fake wandb API.
    fake_wandb.fake_api.add_mock(
        lambda query, idx: {
            "instance": {
                "views_500": {
                    "edges": [
                        {
                            "node": {
                                "type": "runs",
                                "id": "c1233b7003317090ab5e2a75db4ad965",
                                "project": {
                                    "id": "e1233b7003317090ab5e2a75db4ad965",
                                    "name": "test-project-1",
                                },
                            }
                        }
                    ]
                }
            }
        }
    )

    # Compile a project node. Since root_all_reports() lacks a gql_op_output_type,
    # the projects should not have keys.
    projects_node = compile.compile([root_all_reports().project()])[0]
    assert projects_node.type == types.List(wdt.ProjectType)

    # Test that even without serializable instances, the execute works,
    # refining the type to projectTypeWithKeys().
    projects = weave.use(projects_node)
    assert len(projects) == 1

    # Test that keys are restored when converting a literal to a
    # constnode and calling type_of on it.
    names_node = ops.project_ops.name(projects)
    names_node = compile.compile([names_node])[
        0
    ]  # Compile for dispatch to use the correct operation.
    assert names_node.type == types.List(
        TaggedValueType(
            types.TypedDict(
                {
                    "project": wdt.ProjectType.with_keys(
                        {"id": types.String(), "name": types.String()}
                    )
                }
            ),
            types.String(),
        )
    )

    names = weave.use(names_node)
    assert names[0] == "test-project-1"

    # Test that we can convert instances of keyless types to arrow
    # and back to keyless types.

    arrow = projects_node.list_to_arrow()
    awl = weave.use(arrow)
    assert awl[0] == projects[0]


def test_root_artifact_version_gql_propagation():
    node = ops.artifact_version_ops.root_artifact_version(
        "stacey", "mendeleev", "test_res_1fwmcd3q", "v0"
    )
    compiled_node = compile.compile([node])[0]

    expected = types.optional(wdt.ArtifactVersionType.with_keys({"id": types.String()}))
    assert compiled_node.type == expected



import black
import pytest

import weave
from weave.legacy.weave import panels

# IMPORTANT: Do not import other symbols inside of weave
# so that we ensure the produced code only relies on the weave symbol.


@pytest.mark.parametrize(
    "panel, code",
    [
        (1, "1"),
        ("hi", "'hi'"),
        (
            {"a": 1, "b": "hi", "c": True, "d": None},
            "{'a': 1, 'b': 'hi', 'c': True, 'd': None}",
        ),
        (
            panels.Group(),
            "weave.legacy.weave.panels.panel_group.Group()",
        ),
        (
            lambda: panels.Table(
                weave.legacy.weave.ops.range(1, 100, 1).map(
                    lambda row: weave.legacy.weave.ops_primitives.dict.dict_(
                        x=row,
                        y=weave.legacy.weave.ops_primitives.list_.make_list(
                            a=row,
                        ),
                    )
                )
            ),
            lambda: """weave.legacy.weave.panels.panel_table.Table(
            weave.legacy.weave.ops_arrow.list_range.range(1, 100, 1,).map(
                lambda row: weave.legacy.weave.ops_primitives.dict.dict_(
                    x=row,
                    y=weave.legacy.weave.ops_primitives.list_.make_list(
                        a=row,
                    ),
                ),
            ),
        )""",
        ),
        (
            lambda: panels.Plot(
                weave.legacy.weave.ops.range(1, 100, 1).map(
                    lambda row: weave.legacy.weave.ops_primitives.dict.dict_(
                        x=row,
                        y=row**2,
                    )
                )
            ),
            lambda: """weave.legacy.weave.panels.panel_plot.Plot(
            weave.legacy.weave.ops_arrow.list_range.range(1, 100, 1,).map(
                lambda row: weave.legacy.weave.ops_primitives.dict.dict_(
                    x=row,
                    y=row.powBinary(2,),
                ),
            ),
        )""",
        ),
        (
            lambda: panels.Group(
                items={
                    "table": panels.Table(
                        weave.legacy.weave.ops.range(1, 100, 1),
                        columns=[
                            lambda row: row,
                            lambda row: row**2,
                        ],
                    ),
                    "all_rows": lambda table: panels.Plot(
                        table.all_rows(),
                        x=lambda row: row["c_0"],
                        y=lambda row: row["c_1"],
                    ),
                    "derived": lambda table: panels.Group(
                        layoutMode="horizontal",
                        items={
                            "rows": panels.Group(
                                items={
                                    "pinned_rows": panels.Plot(
                                        table.pinned_rows(),
                                        x=lambda row: row["c_0"],
                                        y=lambda row: row["c_1"],
                                    ),
                                    "active_row": table.active_row(),
                                }
                            ),
                            "data": panels.Group(
                                items={
                                    "pinned_data": table.pinned_data(),
                                    "active_data": table.active_data(),
                                }
                            ),
                        },
                    ),
                }
            ),
            lambda: """weave.legacy.weave.panels.panel_group.Group(
    items={
        "table": weave.legacy.weave.panels.panel_table.Table(
            weave.legacy.weave.ops_arrow.list_range.range(1, 100, 1,),
            columns=[
                lambda row: row,
                lambda row: row.powBinary(2,),
            ],
        ),
        "all_rows": lambda table: weave.legacy.weave.panels.panel_plot.Plot(
            table.all_rows(),
            x=lambda row: row["c_0"],
            y=lambda row: row["c_1"],
        ),
        "derived": lambda table, all_rows: weave.legacy.weave.panels.panel_group.Group(
            layoutMode="horizontal",
            items={
                "rows": weave.legacy.weave.panels.panel_group.Group(
                    items={
                        "pinned_rows": weave.legacy.weave.panels.panel_plot.Plot(
                            table.pinned_rows(),
                            x=lambda row: row["c_0"],
                            y=lambda row: row["c_1"],
                        ),
                        "active_row": lambda pinned_rows: table.active_row(),
                    },
                ),
                "data": lambda rows: weave.legacy.weave.panels.panel_group.Group(
                    items={
                        "pinned_data": table.pinned_data(),
                        "active_data": lambda pinned_data: table.active_data(),
                    },
                ),
            },
        ),
    },
)""",
        ),
    ],
)
def test_generic_export(panel, code, consistent_table_col_ids):
    if callable(panel):
        panel = panel()
    if callable(code):
        code = code()
    _test_object_codification(panel, code)


@pytest.mark.skip(reason="Leaving for tim to fix in his branch")
def test_group_case(cereal_csv, consistent_table_col_ids):
    _test_object_codification(
        panels.Group(
            items={
                "plot": panels.Plot(
                    weave.legacy.weave.ops.local_path(cereal_csv).readcsv(),
                    x=lambda row: row["protein"],
                    y=lambda row: row["calories"],
                ),
                "table": lambda plot: panels.Table(
                    plot.selected_rows(),
                    columns=[
                        lambda row: row["c_0"],
                        lambda row: row["c_1"],
                    ],
                ),
            }
        ),
        '''weave.legacy.weave.panels.panel_group.Group(
        items={
            "plot": weave.legacy.weave.panels.panel_plot.Plot(
                weave.legacy.weave.ops.local_path("'''
        + cereal_csv
        + """",).readcsv(),
                x=lambda row: row["protein"],
                y=lambda row: row["calories"],
            ),
            "table": lambda plot: weave.legacy.weave.panels.panel_table.Table(
                plot.selected_rows(),
                columns=[
                    lambda row: row["c_0"],
                    lambda row: row["c_1"],
                ],
            ),
        },
    )""",
    )


@pytest.mark.skip(reason="Leaving for tim to fix in his branch")
def test_plot_case(cereal_csv, consistent_table_col_ids):
    _test_object_codification(
        panels.Plot(
            weave.legacy.weave.ops.local_path(cereal_csv).readcsv(),
            x=lambda row: row["protein"],
            y=lambda row: row["calories"],
        ),
        f"""weave.legacy.weave.panels.panel_plot.Plot(
    weave.legacy.weave.ops.local_path('{cereal_csv}',).readcsv(),
    x=lambda row: row["protein"],
    y=lambda row: row["calories"],
)""",
    )


def test_table_case(cereal_csv, consistent_table_col_ids):
    _test_object_codification(
        panels.Table(
            weave.legacy.weave.ops.local_path(cereal_csv).readcsv(),
            columns=[
                lambda row: row["protein"],
                lambda row: row["calories"],
            ],
        ),
        f"""weave.legacy.weave.panels.panel_table.Table(
    weave.legacy.weave.ops.local_path('{cereal_csv}',).readcsv(),
    columns=[
        lambda row: row["protein"],
        lambda row: row["calories"],
    ],
)""",
    )


def _test_object_codification(panel, code=None):
    panel_code = weave.legacy.weave.codify.object_to_code(panel)

    generated_panel = eval(panel_code)

    panel_json = weave.storage.to_python(panel)
    generated_panel_json = weave.storage.to_python(generated_panel)

    assert panel_json == generated_panel_json

    if code:
        assert panel_code == black.format_str(code, mode=black.FileMode())
